from typing import Sequence, Any


class Options:
    def __init__(self, name: str, values: Sequence):
        self.name = name
        self.values = values

    def validate(self, value: Any) -> bool:
        if value not in self.values:
            valid = "\n- ".join(self.values)
            raise ValueError(
                f"'{value}' is not a valid {self.name}, "
                + f"please choose from one of the following:\n- {valid}"
            )
        return True


correlation_methods = Options(
    "Correlation Method",
    (
        "RoadLink_TOID_TopographicArea_TOID_2",
        "Street_USRN_TopographicArea_TOID_4",
        "BLPU_UPRN_TopographicArea_TOID_5",
        "RoadLink_TOID_Road_TOID_7",
        "RoadLink_TOID_Street_USRN_8",
        "BLPU_UPRN_RoadLink_TOID_9",
        "Road_TOID_Street_USRN_10",
        "BLPU_UPRN_Street_USRN_11",
        "ORRoadLink_GUID_RoadLink_TOID_12",
        "ORRoadNode_GUID_RoadLink_TOID_13",
    ),
)

identifier_types = Options("Identifier Type", ("TOID", "UPRN", "USRN", "GUID"))

feature_types = Options(
    "Feature Type",
    (
        "TopographicArea",
        "RoadLink",
        "Road",
        "BLPU",
        "Street",
        "ORRoadNode",
        "ORRoadLink",
    ),
)



import operator

from osdatahub import Extent
from osdatahub.filters import is_equal, is_less_than, filter_and, filter_or
from pytest import param


def test_get_single_attribute_filter():
    test_variables = "property_name, filter_type, a, expected_result"
    test_data = [
        param(
            "Theme",
            "PropertyIsEqualTo",
            "Building",
            "<ogc:PropertyIsEqualTo><ogc:PropertyName>Theme</ogc:PropertyName><ogc:Literal>Building</ogc:Literal></ogc:PropertyIsEqualTo>",
            id="equal to string",
        ),
        param(
            "Theme",
            "PropertyIsEqualTo",
            "Road",
            "<ogc:PropertyIsEqualTo><ogc:PropertyName>Theme</ogc:PropertyName><ogc:Literal>Road</ogc:Literal></ogc:PropertyIsEqualTo>",
            id="equal to string",
        ),
    ]
    return test_variables, test_data


def test_between():
    test_variables = "property_name, lower, upper, expected_result"
    test_data = [
        param(
            "Width",
            5,
            10,
            f"<ogc:PropertyIsBetween><ogc:PropertyName>Width</ogc:PropertyName><LowerBoundary><ogc:Literal>5</ogc:Literal></LowerBoundary><UpperBoundary><ogc:Literal>10</ogc:Literal></UpperBoundary></ogc:PropertyIsBetween>",
            id="between two ints",
        ),
    ]
    return test_variables, test_data


def test_intersects():
    test_variables = "extent, expected_result"
    test_data = [
        param(
            Extent.from_bbox((0, 0, 1, 1), "EPSG:27700"),
            "<ogc:Intersects>"
            "<ogc:PropertyName>SHAPE</ogc:PropertyName>"
            "<gml:Polygon xmlns:gml='http://www.opengis.net/gml' srsName='EPSG:27700'>"
            "<gml:outerBoundaryIs>"
            "<gml:LinearRing>"
            '<gml:coordinates decimal="." cs="," ts=" ">1.0,0.0 1.0,1.0 0.0,1.0 0.0,0.0 1.0,0.0</gml:coordinates>'
            "</gml:LinearRing>"
            "</gml:outerBoundaryIs>"
            "</gml:Polygon>"
            "</ogc:Intersects>",
            id="EPSG:27700 - bbox",
        ),
        param(
            Extent.from_bbox((0, 0, 1, 1), "EPSG:3857"),
            "<ogc:Intersects>"
            "<ogc:PropertyName>SHAPE</ogc:PropertyName>"
            "<gml:Polygon xmlns:gml='http://www.opengis.net/gml' srsName='EPSG:3857'>"
            "<gml:outerBoundaryIs>"
            "<gml:LinearRing>"
            '<gml:coordinates decimal="." cs="," ts=" ">1.0,0.0 1.0,1.0 0.0,1.0 0.0,0.0 1.0,0.0</gml:coordinates>'
            "</gml:LinearRing>"
            "</gml:outerBoundaryIs>"
            "</gml:Polygon>"
            "</ogc:Intersects>",
            id="EPSG:3857 - bbox",
        ),
        param(
            Extent.from_bbox((0, 0, 1, 1), "epsg:3857"),
            "<ogc:Intersects>"
            "<ogc:PropertyName>SHAPE</ogc:PropertyName>"
            "<gml:Polygon xmlns:gml='http://www.opengis.net/gml' srsName='EPSG:3857'>"
            "<gml:outerBoundaryIs>"
            "<gml:LinearRing>"
            '<gml:coordinates decimal="." cs="," ts=" ">1.0,0.0 1.0,1.0 0.0,1.0 0.0,0.0 1.0,0.0</gml:coordinates>'
            "</gml:LinearRing>"
            "</gml:outerBoundaryIs>"
            "</gml:Polygon>"
            "</ogc:Intersects>",
            id="epsg:3857 - bbox",
        ),
    ]
    return test_variables, test_data


def test_logical():
    test_variables = "filter1, filter2, filter3, op1, op2, op3, expected_result"
    filter_eq = is_equal("eq_property", "eq_value")
    filter_lt = is_less_than("lt_property", 0)
    filter_gt = is_less_than("lt_property", 0)
    test_data = [
        param(
            filter_eq,
            filter_lt,
            filter_gt,
            operator.and_,
            filter_and,
            operator.or_,
            f"<ogc:Or><ogc:And>{filter_eq}{filter_lt}</ogc:And><ogc:And>{filter_lt}{filter_gt}</ogc:And></ogc:Or>",
            id="operator equals string",
        ),
        param(
            filter_eq,
            filter_lt,
            filter_gt,
            operator.or_,
            operator.and_,
            filter_or,
            f"<ogc:Or><ogc:Or>{filter_eq}{filter_lt}</ogc:Or><ogc:And>{filter_lt}{filter_gt}</ogc:And></ogc:Or>",
            id="operator equals string",
        ),
    ]
    return test_variables, test_data



import itertools
from typeguard import TypeCheckError

from osdatahub.DownloadsAPI.downloads_api import _DownloadObj
from pytest import param
from tqdm import tqdm


def generate_product_list_params(file_name, file_format, file_subformat, area, return_downloadobj):
    params = {
        "fileName": file_name,
        "format": file_format,
        "subformat": file_subformat,
        "area": area
    }
    return dict(filter(lambda i: i[1] is not None, params.items()))


def product_list_pass(product_name):
    test_variables = "file_name, file_format, file_subformat, area, return_downloadobj, expected_url, expected_params"
    file_names = ["test_file_name", "test_filename2"]
    file_formats = ["test_file_format"]
    file_subformats = ["test_file_subformat"]
    area = ["GB", "TM"]
    return_downloadobj_values = [True, False]

    permutations = list(itertools.product(file_names, file_formats, file_subformats, area, return_downloadobj_values))
    expected_params = [generate_product_list_params(*p) for p in permutations]
    test_data = [param(*values, f"https://api.os.uk/downloads/v1/products/{product_name}/downloads", expected_param)
                 for values, expected_param in zip(permutations, expected_params)]

    return test_variables, test_data


def product_list_fail():
    test_variables = "file_name, file_format, file_subformat, area, return_downloadobj, expected_result"

    test_data = [
        param(123, None, None, None, False, (TypeError, TypeCheckError)),
        param(None, 123, None, None, False, (TypeError, TypeCheckError)),
        param(None, None, 123, None, False, (TypeError, TypeCheckError)),
        param(None, None, None, 123, False, (TypeError, TypeCheckError)),
        param(None, None, None, "Wrong Area Code", False, (ValueError, TypeCheckError)),
        param(None, None, None, None, "wrong value", (TypeError, TypeCheckError))
    ]
    return test_variables, test_data


def fake_product(i):
    return {
        "md5": f"test_md5{i}",
        "size": f"{i}",
        "url": f"https://testurl{i}.com",
        "format": f"test_format{i}",
        "test_subformat": f"test_subformat{i}",
        "area": "HP",
        "fileName": f"testfile{i}.zip"
    }


def download_pass():
    test_variables = "output_dir, " \
                     "download_multiple, " \
                     "overwrite, " \
                     "processes, " \
                     "product_list_mock_return , " \
                     "downloadobj_mock_return, " \
                     "expected_download_return, " \
                     "download_called_value"

    fake_products = [fake_product(x) for x in range[5]]

    test_data = [
        param(".",
              True,
              True,
              1,
              [_DownloadObj(url="testurl.com", size=100, file_name="testfile.zip")],

              ),
        param(fake_products[0],
              ".",
              False,
              True,
              [".", True, tqdm()]
              ),
        param(fake_products[0],
              "test",
              False,
              True,
              )
    ]

    return test_variables, test_data



from dataclasses import dataclass
from typing import Any, Dict, List, Optional

"""
Structured a bit like Rust does From and Into.
See here: https://doc.rust-lang.org/rust-by-example/conversion/from_into.html
"""


@dataclass
class NGDFeatureCollection:
    """GeoJSON FeatureCollection response from NGD API."""

    type: str
    features: List[Dict[str, Any]]
    numberReturned: int
    links: List[Dict[str, Any]]
    timeStamp: Optional[str] = None
    numberMatched: Optional[int] = None

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "NGDFeatureCollection":
        """Create FeatureCollection from API response dict."""
        return cls(
            type=data.get("type", "FeatureCollection"),
            features=data.get("features", []),
            numberReturned=data.get("numberReturned", 0),
            links=data.get("links", []),
            timeStamp=data.get("timeStamp"),
            numberMatched=data.get("numberMatched"),
        )

    def to_dict(self) -> Dict[str, Any]:
        """Convert to dict for GeoJSON compatibility."""
        result: Dict[str, Any] = {
            "type": self.type,
            "features": self.features,
            "numberReturned": self.numberReturned,
            "links": self.links,
        }
        if self.timeStamp is not None:
            result["timeStamp"] = self.timeStamp
        if self.numberMatched is not None:
            result["numberMatched"] = self.numberMatched
        return result



import pytest
import requests_mock
from osdatahub import Extent

from tests.data import extent_data as data


class TestExtent:
    @pytest.mark.parametrize(*data.test_from_bbox())
    def test_from_bbox(self, bbox, crs, expected_result):
        # Act
        extent = Extent.from_bbox(bbox, crs)

        # Assert
        assert extent == expected_result

    @pytest.mark.parametrize(*data.test_from_radius())
    def test_from_radius(self, centre, radius, crs, expected_result, expected_bbox):
        # Act
        extent = Extent.from_radius(centre, radius, crs)

        # Assert
        assert extent == expected_result
        assert extent.bbox == expected_bbox

    @pytest.mark.parametrize(*data.test_set_crs())
    def test_set_crs(
            self, initital_polygon, initital_crs, updated_crs, expected_result
    ):
        # Arrange
        extent = Extent(initital_polygon, initital_crs)

        # Act
        extent = extent.set_crs(updated_crs)

        # Assert
        assert extent == expected_result

    @pytest.mark.parametrize(*data.test_is_within())
    def test_is_within(self, polygon, crs, bounds, expected_result):
        # Arrange
        extent = Extent(polygon, crs)

        # Act
        within = extent.is_within(bounds)

        # Assert
        assert within == expected_result

    @pytest.mark.parametrize(*data.test_xml_coords())
    def test_xml_coords(self, polygon, crs, expected_result):
        # Arrange
        extent = Extent(polygon, crs)

        # Act
        xml_coords = extent.xml_coords

        # Assert
        assert xml_coords == expected_result

    @pytest.mark.parametrize(*data.test_to_json())
    def test_to_json(self, polygon, crs, expected_result):
        # Arrange
        extent = Extent(polygon, crs)

        # Act
        json = extent.to_json()
        print(json)
        print(expected_result)

        # Assert
        assert json == expected_result


@pytest.mark.parametrize(*data.test_from_ons_code())
def test_from_ons_code(ons_code, ons_mock_response, expected_result):
    # Arrange
    from osdatahub.ons_api import ID_ENDPOINT

    endpoint = f"{ID_ENDPOINT}{ons_code}"

    # Act
    with requests_mock.Mocker() as m:
        m.register_uri("GET", endpoint, text=ons_mock_response)
        extent = Extent.from_ons_code(ons_code)

        # Assert
        assert extent == expected_result


@pytest.mark.parametrize(*data.test_from_ons_code_error())
def test_from_ons_code_error(ons_code, ons_mock_response):
    # Arrange
    from osdatahub.ons_api import ID_ENDPOINT

    endpoint = f"{ID_ENDPOINT}{ons_code}"

    # Act
    with requests_mock.Mocker() as m:
        m.register_uri("GET", endpoint, text=ons_mock_response)
        with pytest.raises(ValueError):
            Extent.from_ons_code(ons_code)



from os import environ

import pytest
from osdatahub.PlacesAPI.places_api import PlacesAPI

from tests.data import places_data as data


class TestFormatFq:
    @pytest.fixture()
    def places(self):
        key = environ.get("OS_API_KEY")
        places = PlacesAPI(key)
        yield places

    @pytest.mark.parametrize(*data.test_format_fq())
    @pytest.mark.usefixtures("places")
    def test_format_fq_pass(
            self, places, classification_codes, logical_states, expected_result
    ):
        # Act
        fq_args = places._PlacesAPI__format_fq(
            classification_code=classification_codes, logical_status_code=logical_states
        )
        # Assert
        assert fq_args == expected_result

    @pytest.mark.parametrize(*data.test_format_fq_errors())
    @pytest.mark.usefixtures("places")
    def test_format_fq_errors(
            self, places, classification_codes, logical_states, expected_result
    ):
        # Act
        with pytest.raises(expected_result):
            fq_args = places._PlacesAPI__format_fq(
                classification_code=classification_codes,
                logical_status_code=logical_states,
            )



from osdatahub import Extent
from osdatahub.filters import intersects, is_between
from pytest import param


def test__params():
    test_variables = "extent, product, filters, expected_result"
    test_data = [
        param(
            Extent.from_bbox((0, 0, 1, 1), "EPSG:27700"),
            "topographic_area",
            [],
            {
                "service": "wfs",
                "version": "2.0.0",
                "request": "GetFeature",
                "outputFormat": "geojson",
                "count": 100,
                "key": "API-KEY",
                "srsName": "EPSG:27700",
                "typeName": "Topography_TopographicArea",
                "filter": '<ogc:Filter><ogc:Intersects><ogc:PropertyName>SHAPE</ogc:PropertyName><gml:Polygon xmlns:gml=\'http://www.opengis.net/gml\' srsName=\'EPSG:27700\'><gml:outerBoundaryIs><gml:LinearRing><gml:coordinates decimal="." cs="," ts=" ">1.0,0.0 1.0,1.0 0.0,1.0 0.0,0.0 1.0,0.0</gml:coordinates></gml:LinearRing></gml:outerBoundaryIs></gml:Polygon></ogc:Intersects></ogc:Filter>',
            },
            id="bbox no filters",
        ),
        param(
            Extent.from_bbox((0, 0, 1, 1), "EPSG:27700"),
            "topographic_area",
            [
                is_between("sample name", 0, 1),
                intersects(Extent.from_bbox((0, 0, 1, 1), "EPSG:27700")),
            ],
            {
                "service": "wfs",
                "version": "2.0.0",
                "request": "GetFeature",
                "outputFormat": "geojson",
                "count": 100,
                "key": "API-KEY",
                "srsName": "EPSG:27700",
                "typeName": "Topography_TopographicArea",
                "filter": '<ogc:Filter><ogc:And><ogc:And><ogc:Intersects><ogc:PropertyName>SHAPE</ogc:PropertyName><gml:Polygon xmlns:gml=\'http://www.opengis.net/gml\' srsName=\'EPSG:27700\'><gml:outerBoundaryIs><gml:LinearRing><gml:coordinates decimal="." cs="," ts=" ">1.0,0.0 1.0,1.0 0.0,1.0 0.0,0.0 1.0,0.0</gml:coordinates></gml:LinearRing></gml:outerBoundaryIs></gml:Polygon></ogc:Intersects><ogc:PropertyIsBetween><ogc:PropertyName>sample name</ogc:PropertyName><LowerBoundary><ogc:Literal>0</ogc:Literal></LowerBoundary><UpperBoundary><ogc:Literal>1</ogc:Literal></UpperBoundary></ogc:PropertyIsBetween></ogc:And><ogc:Intersects><ogc:PropertyName>SHAPE</ogc:PropertyName><gml:Polygon xmlns:gml=\'http://www.opengis.net/gml\' srsName=\'EPSG:27700\'><gml:outerBoundaryIs><gml:LinearRing><gml:coordinates decimal="." cs="," ts=" ">1.0,0.0 1.0,1.0 0.0,1.0 0.0,0.0 1.0,0.0</gml:coordinates></gml:LinearRing></gml:outerBoundaryIs></gml:Polygon></ogc:Intersects></ogc:And></ogc:Filter>',
            },
            id="bbox with filters is_between intersects",
        ),
        param(
            Extent.from_bbox((0, 0, 1, 1), "EPSG:27700"),
            "Topography_TopographicArea",
            [
                is_between("sample name", 0, 1),
                intersects(Extent.from_bbox((0, 0, 1, 1), "EPSG:27700")),
            ],
            {
                "service": "wfs",
                "version": "2.0.0",
                "request": "GetFeature",
                "outputFormat": "geojson",
                "count": 100,
                "key": "API-KEY",
                "srsName": "EPSG:27700",
                "typeName": "Topography_TopographicArea",
                "filter": '<ogc:Filter><ogc:And><ogc:And><ogc:Intersects><ogc:PropertyName>SHAPE</ogc:PropertyName><gml:Polygon xmlns:gml=\'http://www.opengis.net/gml\' srsName=\'EPSG:27700\'><gml:outerBoundaryIs><gml:LinearRing><gml:coordinates decimal="." cs="," ts=" ">1.0,0.0 1.0,1.0 0.0,1.0 0.0,0.0 1.0,0.0</gml:coordinates></gml:LinearRing></gml:outerBoundaryIs></gml:Polygon></ogc:Intersects><ogc:PropertyIsBetween><ogc:PropertyName>sample name</ogc:PropertyName><LowerBoundary><ogc:Literal>0</ogc:Literal></LowerBoundary><UpperBoundary><ogc:Literal>1</ogc:Literal></UpperBoundary></ogc:PropertyIsBetween></ogc:And><ogc:Intersects><ogc:PropertyName>SHAPE</ogc:PropertyName><gml:Polygon xmlns:gml=\'http://www.opengis.net/gml\' srsName=\'EPSG:27700\'><gml:outerBoundaryIs><gml:LinearRing><gml:coordinates decimal="." cs="," ts=" ">1.0,0.0 1.0,1.0 0.0,1.0 0.0,0.0 1.0,0.0</gml:coordinates></gml:LinearRing></gml:outerBoundaryIs></gml:Polygon></ogc:Intersects></ogc:And></ogc:Filter>',
            },
            id="API style product name",
        ),
    ]
    return test_variables, test_data


def test_request_params():
    test_variables = "extent, product, filters, limit, expected_url, expected_params"
    test_data = [
        param(
            Extent.from_bbox((0, 0, 1, 1), "EPSG:27700"),
            "topographic_area",
            [],
            100,
            "https://api.os.uk/features/v1/wfs",
            {
                "service": "wfs",
                "version": "2.0.0",
                "request": "GetFeature",
                "outputFormat": "geojson",
                "key": "API-KEY",
                "srsName": "EPSG:27700",
                "typeName": "Topography_TopographicArea",
                "filter": '<ogc:Filter><ogc:Intersects><ogc:PropertyName>SHAPE</ogc:PropertyName><gml:Polygon xmlns:gml=\'http://www.opengis.net/gml\' srsName=\'EPSG:27700\'><gml:outerBoundaryIs><gml:LinearRing><gml:coordinates decimal="." cs="," ts=" ">1.0,0.0 1.0,1.0 0.0,1.0 0.0,0.0 1.0,0.0</gml:coordinates></gml:LinearRing></gml:outerBoundaryIs></gml:Polygon></ogc:Intersects></ogc:Filter>',
                "count": 100,
                "startIndex": 0
            },
            id="bbox no filters - limit=100",
        ),
        param(
            Extent.from_bbox((0, 0, 1, 1), "EPSG:27700"),
            "topographic_area",
            [],
            50,
            "https://api.os.uk/features/v1/wfs",
            {
                "service": "wfs",
                "version": "2.0.0",
                "request": "GetFeature",
                "outputFormat": "geojson",
                "key": "API-KEY",
                "srsName": "EPSG:27700",
                "typeName": "Topography_TopographicArea",
                "filter": '<ogc:Filter><ogc:Intersects><ogc:PropertyName>SHAPE</ogc:PropertyName><gml:Polygon xmlns:gml=\'http://www.opengis.net/gml\' srsName=\'EPSG:27700\'><gml:outerBoundaryIs><gml:LinearRing><gml:coordinates decimal="." cs="," ts=" ">1.0,0.0 1.0,1.0 0.0,1.0 0.0,0.0 1.0,0.0</gml:coordinates></gml:LinearRing></gml:outerBoundaryIs></gml:Polygon></ogc:Intersects></ogc:Filter>',
                "count": 50,
                "startIndex": 0
            },
            id="bbox no filters - limit=50",
        ),
    ]
    return test_variables, test_data


def test_query():
    test_variables = "product, extent, limit, expected_count"
    test_data = [
        param(
            "topographic_area",
            Extent.from_bbox((600000, 310200, 600900, 310900), "EPSG:27700"),
            100,
            100,
            id="bbox, topographic_area, limit=100",
        ),
        param(
            "topographic_area",
            Extent.from_bbox((600000, 310200, 600900, 310900), "EPSG:27700"),
            12,
            12,
            id="bbox, topographic_area, limit=12",
        ),
        param(
            "Topography_TopographicArea",
            Extent.from_bbox((600000, 310200, 600900, 310900), "EPSG:27700"),
            12,
            12,
            id="bbox, Topography_TopographicArea, limit=12",
        ),
        param(
            "topographic_area",
            Extent.from_bbox((600000, 310200, 600900, 310900), "EPSG:27700"),
            200,
            200,
            id="bbox, topographic_area, limit=200",
        ),
    ]
    return test_variables, test_data



from collections import namedtuple
from collections.abc import Iterable
from typing import Union

Local_Type = namedtuple("LocalType", "Type local_type")

LOCAL_TYPE = {
    "airfield": Local_Type("transportNetwork", "Airfield"),
    "airport": Local_Type("transportNetwork", "Airport"),
    "bay": Local_Type("hydrography", "Bay"),
    "beach": Local_Type("landcover", "Beach"),
    "bus_station": Local_Type("transportNetwork", "Bus_Station"),
    "channel": Local_Type("hydrography", "Channel"),
    "chemical_works": Local_Type("other", "Chemical_Works"),
    "cirque_or_hollow": Local_Type("landform", "Cirque_Or_Hollow"),
    "city": Local_Type("populatedPlace", "City"),
    "cliff_or_slope": Local_Type("landform", "Cliff_Or_Slope"),
    "coach_station": Local_Type("transportNetwork", "Coach_Station"),
    "coastal_headland": Local_Type("landform", "Coastal_Headland"),
    "electricity_distribution": Local_Type("other", "Electricity_Distribution"),
    "electricity_production": Local_Type("other", "Electricity_Production"),
    "estuary": Local_Type("hydrography", "Estuary"),
    "further_education": Local_Type("other", "Further_Education"),
    "gas_distribution_or_storage": Local_Type("other", "Gas_Distribution_or_Storage"),
    "group_of_islands": Local_Type("landform", "Group_Of_Islands"),
    "hamlet": Local_Type("populatedPlace", "Hamlet"),
    "harbour": Local_Type("transportNetwork", "Harbour"),
    "helicopter_station": Local_Type("transportNetwork", "Helicopter_Station"),
    "heliport": Local_Type("transportNetwork", "Heliport"),
    "higher_or_university_education": Local_Type("other", "Higher_or_University_Education"),
    "hill_or_mountain": Local_Type("landform", "Hill_Or_Mountain"),
    "hill_or_mountain_ranges": Local_Type("landform", "Hill_Or_Mountain_Ranges"),
    "hospice": Local_Type("other", "Hospice"),
    "hospital": Local_Type("other", "Hospital"),
    "inland_water": Local_Type("hydrography", "Inland_Water"),
    "island": Local_Type("landform", "Island"),
    "medical_care_accommodation": Local_Type("other", "Medical_Care_Accommodation"),
    "named_road": Local_Type("transportNetwork", "Named_Road"),
    "non_state_primary_education": Local_Type("other", "Non_State_Primary_Education"),
    "non_state_secondary_education": Local_Type("other", "Non_State_Secondary_Education"),
    "numbered_road": Local_Type("transportNetwork", "Numbered_Road"),
    "oil_distribution_or_storage": Local_Type("other", "Oil_Distribution_or_Storage"),
    "oil_refining": Local_Type("other", "Oil_Refining"),
    "oil_terminal": Local_Type("other", "Oil_Terminal"),
    "other_coastal_landform": Local_Type("landform", "Other_Coastal_Landform"),
    "other_landcover": Local_Type("landcover", "Other_Landcover"),
    "other_landform": Local_Type("landform", "Other_Landform"),
    "other_settlement": Local_Type("populatedPlace", "Other_Settlement"),
    "passenger_ferry_terminal": Local_Type("transportNetwork", "Passenger_Ferry_Terminal"),
    "port_consisting_of_docks_and_nautical_berthing": Local_Type("transportNetwork",
                                                                 "Port_Consisting_of_Docks_and_Nautical_Berthing"),
    "postcode": Local_Type("other", "Postcode"),
    "primary_education": Local_Type("other", "Primary_Education"),
    "railway": Local_Type("transportNetwork", "Railway"),
    "railway_station": Local_Type("transportNetwork", "Railway_Station"),
    "road_user_services": Local_Type("transportNetwork", "Road_User_Services"),
    "sea": Local_Type("hydrography", "Sea"),
    "secondary_education": Local_Type("other", "Secondary_Education"),
    "section_of_named_road": Local_Type("transportNetwork", "Section_Of_Named_Road"),
    "section_of_numbered_road": Local_Type("transportNetwork", "Section_Of_Numbered_Road"),
    "special_needs_education": Local_Type("other", "Special_Needs_Education"),
    "spot_height": Local_Type("landform", "Spot_Height"),
    "suburban_area": Local_Type("populatedPlace", "Suburban_Area"),
    "tidal_water": Local_Type("hydrography", "Tidal_Water"),
    "town": Local_Type("populatedPlace", "Town"),
    "tramway": Local_Type("transportNetwork", "Tramway"),
    "urban_greenspace": Local_Type("landcover", "Urban_Greenspace"),
    "valley": Local_Type("landform", "Valley"),
    "vehicular_ferry_terminal": Local_Type("transportNetwork", "Vehicular_Ferry_Terminal"),
    "vehicular_rail_terminal": Local_Type("transportNetwork", "Vehicular_Rail_Terminal"),
    "village": Local_Type("populatedPlace", "Village"),
    "waterfall": Local_Type("hydrography", "Waterfall"),
    "wetland": Local_Type("landcover", "Wetland"),
    "woodland_or_forest": Local_Type("landcover", "Woodland_Or_Forest")
}


def validate_local_type(local_type: Union[str, Iterable]) -> set:
    local_type = {local_type} if isinstance(local_type, str) else set(local_type)
    return local_type - set(LOCAL_TYPE.keys())


def get_local_type(local_type_name: str) -> str:
    return LOCAL_TYPE[local_type_name].local_type



import json
import warnings

import requests
from geojson import FeatureCollection
from typeguard import typechecked

import osdatahub
from osdatahub.extent import Extent
from osdatahub.errors import raise_http_error
from osdatahub.FeaturesAPI.feature_products import (get_product,
                                                    validate_product_name)
from osdatahub.filters import Filter
from osdatahub.grow_list import GrowList
from osdatahub.spatial_filter_types import SpatialFilterTypes
from osdatahub.utils import features_to_geojson, is_new_api

class FeaturesAPI:
    """Main class for querying the OS Features API (https://osdatahub.os.uk/docs/wfs/overview)

    Args:
        key (str): A valid OS API Key. Get a free key here - https://osdatahub.os.uk/
        product_name (str): A valid OS product
        extent (Extent): The geographical extent of your query
        spatial_filter_type (str): Set the default spatial filter operation (defaults to "intersects")

    Example::

        from osdatahub import FeaturesAPI, Extent
        from os import environ

        key = environ.get("OS_API_KEY")
        extent = Extent.from_bbox((600000, 310200, 600900, 310900), "EPSG:27700")
        features = FeaturesAPI(key, "zoomstack_local_buildings", extent)
        results = features.query(limit=50)
    """

    ENDPOINT = r"https://api.os.uk/features/v1/wfs"

    DEFAULTS = {
        "service": "wfs",
        "version": "2.0.0",
        "request": "GetFeature",
        "outputFormat": "geojson",
        "count": 100,
    }

    def __init__(self, key: str, product_name: str, extent: Extent, spatial_filter_type: str = "intersects"):
        self.key: str = key
        self.new_api: bool = False
        self.product: str = product_name
        self.extent: Extent = extent
        self.filters: list = []
        self.__spatial_filter = SpatialFilterTypes.get(spatial_filter_type)

    @property
    def extent(self):
        return self.__extent

    @extent.setter
    def extent(self, o: object):
        if isinstance(o, Extent):
            self.__extent = o
        else:
            o_type = type(o)
            raise TypeError(
                f'type of argument "extent" must be '
                f"osdatahub.extent.Extent; got {o_type} instead"
            )

    @property
    def product(self):
        return self.__product

    @product.setter
    def product(self, product_name: str):
        product_name = validate_product_name(product_name)
        self.__product = get_product(product_name, self.new_api)
        self.__product_name = product_name

    @property
    def xml_filter(self):
        return self.__construct_filter()

    @typechecked
    def query(self, limit: int = 100) -> FeatureCollection:
        """Run a query of the OS Features API

        Args:
            limit (int, optional): The maximum number of features to return.
                Defaults to 100.

        Returns:
            FeatureCollection: The results of the query in GeoJSON format
        """

        params = self.__params
        data = GrowList()
        n_required = min(limit, 100)
        if "srsName='EPSG:4326'" in params["filter"]:
            warnings.warn("The features API does not support EPSG:4326 and will return an empty features list.")

        try:
            while n_required > 0 and data.grown:
                params.update({"count": n_required, "startIndex": len(data)})
                response = osdatahub.get(self.ENDPOINT, params=params, proxies=osdatahub.get_proxies())
                resp_json = response.json()
                if "fault" in resp_json:
                    raise_http_error(response)
                if is_new_api(resp_json):
                    self.new_api = True
                    self.product = self.__product_name
                data.extend(resp_json["features"])
                n_required = min(100, limit - len(data))
        except json.decoder.JSONDecodeError:
            raise_http_error(response)

        if data and not is_new_api(data):
            warnings.warn("The OS Data Hub has updated the Features API, fixing some important bugs and adding some "
                          "new properties to all responses.\nTo access these features, consider regenerating your API "
                          "key in the OS Data Hub API dashboard.\nMore information about the update can be found at"
                          "osdatahub.os.uk.", DeprecationWarning)
        return features_to_geojson(data.values, self.product.geometry,
                                   self.extent.crs)

    def __construct_filter(self) -> str:
        filter_body = self.__spatial_filter(self.extent)
        for _filter in self.filters:
            filter_body += _filter
            filter_body = f"<ogc:And>{filter_body}</ogc:And>"
        return f"<ogc:Filter>{filter_body}</ogc:Filter>"

    @property
    def __params(self) -> dict:
        return {
            **self.DEFAULTS,
            "key": self.key,
            "srsName": self.extent.crs,
            "typeName": self.product.name,
            "filter": self.__construct_filter(),
        }
    
    @typechecked
    def add_filters(self, *xml_filters: Filter) -> None:
        """Add XML filter strings to the final query

        Args:
            xml_filters (str): Valid OGC XML filter objects
        """
        
        self.filters.extend(xml_filters)


if __name__ == "__main__":
    from os import environ

    from osdatahub import Extent, FeaturesAPI

    key = environ.get("OS_API_KEY")

    bbox = (-1.5339625, 53.787380, -1.5266132, 53.794103)
    extent = Extent.from_bbox(bbox, "EPSG:4326")

    product = "Zoomstack_Sites"
    features = FeaturesAPI(key, product, extent, spatial_filter_type="equals")
    results = features.query()

    print(len(results["features"]))



from pathlib import Path
from typing import Union

import requests
from typeguard import typechecked

import osdatahub
from osdatahub.codes import AREA_CODES

from .downloads_api import _DownloadObj, _DownloadsAPIBase


class OpenDataDownload(_DownloadsAPIBase):
    """
    Main class for downloading OS OpenData products
    (https://osdatahub.os.uk/docs/downloads/technicalSpecification#/OS%20OpenData)

    Args:
        product_id (str): Valid ID for a Downloads API Product or DataPackage.
    """
    _ENDPOINT = _DownloadsAPIBase._ENDPOINT + "products"

    # TODO: change name
    @typechecked
    def product_list(self, 
                     file_name: Union[str, None] = None, 
                     file_format: Union[str, None] = None, 
                     file_subformat: Union[str, None] = None,
                     area: Union[str, None] = None, 
                     return_downloadobj: bool = False) -> Union[list, dict]:
        """
        Returns a list of possible downloads for a specific OS OpenData Product based on given filters

        Args:
            file_name (str, optional): Filter the list of downloads to only include those with this file name
            file_format (str, optional): Filter the list of downloads to only include those with this format
            file_subformat (str, optional): Filter the list of downloads to only include those with this subformat
            area (str, optional) Filter the list of downloads to only include those that cover this area. Available
                values can be found `here <https://osdatahub.os.uk/docs/downloads/technicalSpecification#/OS%20OpenData/get_products__productId__downloads>`_
            return_downloadobj (bool, optional): Returns the downloadable files as _DownloadObj objects instead of as a
                json. Defaults to False

        Returns:
            List of downloadable files from Downloads API
        """
        params = {}
        if area and area not in AREA_CODES:
            raise ValueError(f"Invalid argument \"area\". Had value {area} but must be one of {AREA_CODES}")

        if file_name:
            params.update({"fileName": file_name})
        if file_format:
            params.update({"format": file_format})
        if file_subformat:
            params.update({"subformat": file_subformat})
        if area:
            params.update({"area": area})

        response = osdatahub.get(url=self._endpoint(f"{self._id}/downloads"), params=params, proxies=osdatahub.get_proxies())
        response.raise_for_status()
        if return_downloadobj:
            return [_DownloadObj(url=download["url"], file_name=download["fileName"], size=download["size"])
                    for download in response.json()]
        else:
            return response.json()

    def download(self, 
                 output_dir: Union[str, Path] = ".",
                 file_name: Union[str, None] = None,
                 file_format: Union[str, None] = None,
                 file_subformat: Union[str, None] = None,
                 area: Union[str, None] = None,
                 download_multiple: bool = False,
                 overwrite: bool = False,
                 processes: Union[int, None] = None) -> list:
        """
        Downloads Product files to your local machine

        Args:
            output_dir (Union[str, Path], optional): the path where the downloaded files will be saved. Defaults to
                current working directory
            file_name (str, optional): name of the file(s) to download
            file_format (str, optional): format of the file(s) to download
            file_subformat (str, optional): subformat of the file(s) to download
            area (str, optional): The area that the file must cover. Available values can be found in
                osdatahub.codes.AREA_CODES or on the `OS Data Hub website <https://osdatahub.os.uk/docs/downloads/technicalSpecification#/OS%20OpenData/get_products__productId__downloads>`_
            download_multiple (bool, optional): whether to download multiple files if multiple products are within your
                search criteria. Defaults to False
            overwrite (bool, optional): whether to overwrite existing files. Defaults to False
            processes (int, optional): Number of processes with which to download multiple files. Only relevant if
                multiple files will be downloaded (and download_multiple is set to True)
        """
        download_list = self.product_list(file_name=file_name, file_format=file_format, file_subformat=file_subformat,
                                          area=area, return_downloadobj=True)
        return super()._download(download_list=download_list,
                                 output_dir=output_dir,
                                 overwrite=overwrite,
                                 download_multiple=download_multiple,
                                 processes=processes)



from typing import Union

class GrowList:
    """
    GrowList is a convenience class that behaves similarly to a normal
    list, except that it stores its length changes whenever it is extended
    with the extend() function.
    """

    def __init__(self, values: Union[list, None] = None):
        self.values = values if values else []
        self.size = [len(self.values)]

    def __bool__(self):
        return bool(self.values)

    def __len__(self):
        return len(self.values)

    def __repr__(self):
        return repr(self.values)

    def __iter__(self):
        return iter(self.values)

    @property
    def grown(self):
        """Checks whether the GrowList has been expanded, or whether it has
        been extended with an empty list of values.

        Returns:
            bool: True suggests that the GrowList has recently grown, False
            suggests that it has been extended with an empty list
        """
        if len(self.size) == 1 or self.size[-1] > self.size[-2]:
            return True
        return False

    def __log_size(self):
        self.size.append(len(self.values))

    def extend(self, values: list):
        """Adds values onto the end of the GrowList

        Args:
            values (list): list of values to be added
        """
        self.values.extend(values)
        self.__log_size()



import pytest
from osdatahub.filters import is_between, single_attribute_filter, intersects

from tests.data import filters_data as data


class TestFiltering:
    @pytest.mark.parametrize(*data.test_get_single_attribute_filter())
    def test_get_single_attribute_filter(
            self, property_name, filter_type, a, expected_result
    ):
        # Act
        extent_filter = single_attribute_filter(property_name, filter_type, a)

        # Assert
        assert extent_filter == expected_result

    @pytest.mark.parametrize(*data.test_between())
    def test_between(self, property_name, lower, upper, expected_result):
        # Act
        extent_filter = is_between(property_name, lower, upper)

        # Assert
        assert extent_filter == expected_result

    @pytest.mark.parametrize(*data.test_intersects())
    def test_intersects(self, extent, expected_result):
        # Act
        extent_filter = intersects(extent)

        # Assert
        assert extent_filter == expected_result

    @pytest.mark.parametrize(*data.test_logical())
    def test_logical(self, filter1, filter2, filter3, op1, op2, op3, expected_result):
        # Act
        result = op3(op1(filter1, filter2), op2(filter2, filter3))

        # Assert
        assert result == expected_result



"""
This requests wrapper is included to handle incomplete responses from the api.
Information and inspiration from https://blog.petrzemek.net/2018/04/22/on-incomplete-http-reads-and-the-requests-library-in-python/
"""

import requests

_USER_AGENT_TAG = 'osdatahub-python'

def check_length(func):
    """
    Decorator function that checks if the response content length is as expected.
    """
    def wrapper(*args, **kwargs):
        """
        Wrapper function that performs the content length check.

        Returns:
            The response object if the content length check passes.

        Raises:
            IOError: If the content length check fails.
        """
        response = func(*args, **kwargs)

        expected_length = response.headers.get('Content-Length')

        if expected_length is None:
            return response

        actual_length = response.raw.tell()
        expected_length = int(expected_length)
        if actual_length < expected_length:
            deficit = expected_length - actual_length
            raise IOError(
                f'incomplete read ({actual_length} bytes read, {deficit} more expected)'
            )

        return response
    return wrapper

def add_user_agent_tag(kwargs):
    """
    Adds a User-Agent header to the request so that it matches the USER AGENT TAG.
    """

    if kwargs.get('headers') is None:
        kwargs['headers'] = {}

    kwargs['headers'].update({'User-Agent': _USER_AGENT_TAG})

    return kwargs

@check_length
def get(*args, **kwargs):
    """
    Sends a GET request to the specified URL.

    Args:
        *args: Positional arguments to be passed to the requests.get function.
        **kwargs: Keyword arguments to be passed to the requests.get function.

    Returns:
        The response object if the content length check passes.

    Raises:
        IOError: If the content length check fails.
    """
    kwargs = add_user_agent_tag(kwargs)
    return requests.get(*args, **kwargs)


@check_length
def post(*args, **kwargs):
    """
    Sends a POST request to the specified URL.

    Args:
        *args: Positional arguments to be passed to the requests.post function.
        **kwargs: Keyword arguments to be passed to the requests.post function.

    Returns:
        The response object if the content length check passes.

    Raises:
        IOError: If the content length check fails.
    """
    kwargs = add_user_agent_tag(kwargs)
    return requests.post(*args, **kwargs)



import asyncio
import time


class RateLimiter:
    """
    This class provides two levels of rate limiting:

    1. Concurrent request limiting via asyncio.Semaphore
    2. Time-based delay between requests to prevent burst traffic

    Args:
        max_concurrent: Maximum number of concurrent requests (default: 5)
        request_delay: Minimum delay in seconds between requests (default: 0.02)

    Example::

        limiter = RateLimiter(max_concurrent=5, request_delay=0.3)
        async with limiter:
            await make_request()
    """

    def __init__(self, max_concurrent: int = 5, request_delay: float = 0.02) -> None:
        self._semaphore: asyncio.Semaphore = asyncio.Semaphore(max_concurrent)
        self._request_delay: float = request_delay
        self._last_request_time: float = 0.0
        self._lock: asyncio.Lock = asyncio.Lock()

    async def __aenter__(self) -> "RateLimiter":
        """Acquire semaphore and enforce delay between requests."""
        await self._semaphore.acquire()
        await self._enforce_delay()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        """Release the semaphore."""
        self._semaphore.release()

    async def _enforce_delay(self) -> None:
        """Ensure minimum delay between requests."""
        async with self._lock:
            now = time.monotonic()
            elapsed = now - self._last_request_time
            if elapsed < self._request_delay:
                await asyncio.sleep(self._request_delay - elapsed)
            self._last_request_time = time.monotonic()

    @property
    def max_concurrent(self) -> int:
        """Return the maximum number of concurrent requests allowed."""
        return self._semaphore._value

    @property
    def request_delay(self) -> float:
        """Return the minimum delay between requests."""
        return self._request_delay



import pytest
from osdatahub.utils import address_to_feature, validate_in_range

from tests.data import utils_data as data


class TestAddressToFeature:
    @pytest.mark.parametrize(*data.test_convert_address_to_feature())
    def test_convert_address_to_feature(self, address, crs, expected_result):
        # Act
        address_feature = address_to_feature(address, crs)
        # Assert
        assert address_feature == expected_result

    @pytest.mark.parametrize(*data.test_address_to_feature_error())
    def test_address_to_feature_error(self, address, crs):
        # Act
        with pytest.raises(ValueError):
            address_feature = address_to_feature(address, crs)

    def test_validate_in_range(self):
        # Assert
        assert validate_in_range(10, 5, 15) == 10

    def test_validate_in_range_fail(self, ):
        # Act
        with pytest.raises(ValueError) as error:
            validate_in_range(80, 5, 15) == 10

        # Assert
        assert str(error.value) == "Value should be between 5 and 15, got 80."



from osdatahub.NGD.crs import EPSG
from pytest import param


def test_ngd_crs():
    test_variables = "crs, valid_crs, expected_result"
    test_data = [
        param("epsg:4326", EPSG, EPSG["epsg:4326"]),
        param("EpSg:4326", EPSG, EPSG["epsg:4326"]),
        param("EPSG:27700", EPSG, EPSG["epsg:27700"]),
        param("EPSG:7405", EPSG, EPSG["epsg:7405"]),
        param("EPSG:3857", EPSG, EPSG["epsg:3857"]),
        param("crs84", EPSG, EPSG["crs84"]),
        param("CRS84", EPSG, EPSG["crs84"]),
        param(4326, EPSG, EPSG["epsg:4326"]),
        param(3857, EPSG, EPSG["epsg:3857"]),
        param(27700, EPSG, EPSG["epsg:27700"]),
        param(7405, EPSG, EPSG["epsg:7405"]),
        param(7405, ["epsg:7405"], EPSG["epsg:7405"]),
        param(3857, ("epsg:7405", "epsg:3857"), EPSG["epsg:3857"]),
        param("http://www.opengis.net/def/crs/EPSG/0/4326/", EPSG, EPSG["epsg:4326"]),
        param("www.opengis.net/def/crs/EPSG/0/4326", EPSG, EPSG["epsg:4326"]),
        param("http://www.opengis.net/def/crs/OGC/1.3/crs84", EPSG, EPSG["crs84"]),
        param("HTTPS://www.opengis.net/def/crs/OGC/1.3/CRS84", EPSG, EPSG["crs84"]),
    ]
    return test_variables, test_data


def test_ngd_crs_fail():
    test_variables = "crs, valid_crs, expected_result"
    test_data = [
        param("TEST", EPSG, ValueError),
        param(4326, ("EPSG:7405", "EPSG:3857"), ValueError)
    ]
    return test_variables, test_data


def test_merge_geojsons_pass():
    test_variables = "crs, expected_result"
    test_data = [
        param("TEST", ValueError)
    ]
    return test_variables, test_data



from typing import Union
# import re

EPSG = {
    "epsg:4326": "http://www.opengis.net/def/crs/EPSG/0/4326",
    "epsg:27700": "http://www.opengis.net/def/crs/EPSG/0/27700",
    "epsg:3857": "http://www.opengis.net/def/crs/EPSG/0/3857",
    "epsg:7405": "http://www.opengis.net/def/crs/EPSG/0/7405",
    "crs84": "http://www.opengis.net/def/crs/OGC/1.3/CRS84"
}

def get_crs(crs: Union[str,int] = None, valid_crs: Union[list, tuple] = EPSG.keys()) -> str:
    """Return a CRS in the correct format for the NGD API (URL).  Checks the validity of the CRS
    according to the API and the valid_crs parameter.  

    Args:
        crs (Union[str,int]): A crs code as a string, int or url
        valid_crs (Union[list, tuple]): a list or tuple of valid crs, this is an internal parameter not set by user.

    Returns (str): A url as a string in the correct format for the NGD API

    Example::
    getcrs(crs="epsg:4326", valid_crs=("epsg:4326", "epsg:27700", "epsg:3857", "crs84"))
    """
    if not set(valid_crs).issubset(set(EPSG)):
        raise ValueError(f"`valid_crs` parameter is not valid. Must be an iterable containing only {EPSG.keys()} but "
                         f"had value {valid_crs}")

    if isinstance(crs, int):
        crs = f"epsg:{crs}"

    crs = crs.lower()
    if crs in EPSG and crs in valid_crs:
        return EPSG[crs]  
    else:
        #Checks if the url provided is or closely corresponds to a valid crs url as defined by NGD API spec
        for key, url in EPSG.items():
            if url[7:].casefold() in crs.casefold() and len(crs)<50:
                crs = key
                return EPSG[crs]

    raise ValueError(f'Unknown CRS. Must be in format "epsg:4326", "EPSG:4326", 4326, or '
                     f'"http://www.opengis.net/def/crs/EPSG/0/4326". See documentation for available crs.')



from pytest import param
from typeguard import TypeCheckError

def test_format_fq():
    test_variables = "classification_codes, logical_states, expected_result"
    test_data = [
        param("ABC", None, ["classification_code:ABC"]),
        param(
            ("ABC", "DEF"), None, ["classification_code:ABC classification_code:DEF"]
        ),
        param(
            ["ABC", "DEF"], None, ["classification_code:ABC classification_code:DEF"]
        ),
        param((), None, []),
        param(None, "5", ["logical_status_code:5"]),
        param(None, 5, ["logical_status_code:5"]),
        param("ABC", "5", ["classification_code:ABC", "logical_status_code:5"]),
        param(
            ("ABC", "DEF"),
            5,
            [
                "classification_code:ABC classification_code:DEF",
                "logical_status_code:5",
            ],
        ),
    ]
    return test_variables, test_data


def test_format_fq_errors():
    test_variables = "classification_codes, logical_states, expected_result"
    test_data = [
        param(1, None, (TypeError, TypeCheckError)),
        param(None, "ABC", (TypeError, TypeCheckError)),
        param(None, ("1", "2"), (TypeError, TypeCheckError)),
    ]
    return test_variables, test_data



import os
import unittest.mock as mock

import pytest
from osdatahub import FeaturesAPI

from tests.data import features_api_data as data

API_KEY = os.environ.get("OSDATAHUB_TEST_KEY")


class TestFeaturesAPI:
    @pytest.mark.parametrize(*data.test__params())
    def test__params(self, extent, product, filters, expected_result):
        # Arrange
        features_api = FeaturesAPI("API-KEY", product, extent)
        for filter in filters:
            features_api.add_filters(filter)

        # Act
        params = features_api._FeaturesAPI__params

        # Assert
        assert params == expected_result

    @pytest.mark.parametrize(*data.test_request_params())
    @mock.patch('osdatahub.get')
    def test_request_params(self, request_mocked, extent, product, filters,
                            limit, expected_url, expected_params):
        # Arrange
        request_mocked.return_value.configure_mock(json=lambda: {"features": []})
        features_api = FeaturesAPI("API-KEY", product, extent)
        for filter in filters:
            features_api.add_filters(filter)

        # Act
        features_api.query(limit=limit)

        # Assert
        request_mocked.assert_called_with(expected_url,
                                          params=expected_params,
                                          proxies={})

    @pytest.mark.skipif(not API_KEY, reason="Test API key not available")
    @pytest.mark.parametrize(*data.test_query())
    def test_query(self, product, extent, limit, expected_count):
        # Arrange
        features_api = FeaturesAPI(API_KEY, product, extent)

        # Act
        results = features_api.query(limit=limit)

        # Assert
        assert len(results["features"]) == expected_count



from pytest import param


def test_convert_address_to_feature():
    test_variables = "address, crs, expected_result"
    test_data = [
        param(
            {
                "UPRN": "10094212630",
                "UDPRN": "55819024",
                "ADDRESS": "UNIT D4, ADANAC PARK, ADANAC DRIVE, NURSLING, SOUTHAMPTON, SO16 0BT",
                "SUB_BUILDING_NAME": "UNIT D4",
                "BUILDING_NAME": "ADANAC PARK",
                "THOROUGHFARE_NAME": "ADANAC DRIVE",
                "DEPENDENT_LOCALITY": "NURSLING",
                "POST_TOWN": "SOUTHAMPTON",
                "POSTCODE": "SO16 0BT",
                "RPC": "2",
                "X_COORDINATE": 437230.0,
                "Y_COORDINATE": 115775.0,
                "LNG": -163806.07,
                "LAT": 6610726.44,
                "STATUS": "HISTORICAL",
                "LOGICAL_STATUS_CODE": "8",
                "CLASSIFICATION_CODE": "CI01",
                "CLASSIFICATION_CODE_DESCRIPTION": "Factory/Manufacturing",
                "LOCAL_CUSTODIAN_CODE": 1760,
                "LOCAL_CUSTODIAN_CODE_DESCRIPTION": "TEST VALLEY",
                "POSTAL_ADDRESS_CODE": "D",
                "POSTAL_ADDRESS_CODE_DESCRIPTION": "A record which is linked to PAF",
                "BLPU_STATE_CODE": "null",
                "BLPU_STATE_CODE_DESCRIPTION": "Unknown/Not applicable",
                "TOPOGRAPHY_LAYER_TOID": "osgb5000005278503035",
                "PARENT_UPRN": "10094212626",
                "LAST_UPDATE_DATE": "01/09/2021",
                "ENTRY_DATE": "20/08/2019",
                "LANGUAGE": "EN",
                "MATCH": 0.4,
                "MATCH_DESCRIPTION": "NO MATCH",
            },
            "epsg:3857",
            {
                "type": "Feature",
                "geometry": {"type": "Point", "coordinates": [-163806.07, 6610726.44]},
                "properties": {
                    "UPRN": "10094212630",
                    "UDPRN": "55819024",
                    "ADDRESS": "UNIT D4, ADANAC PARK, ADANAC DRIVE, NURSLING, SOUTHAMPTON, SO16 0BT",
                    "SUB_BUILDING_NAME": "UNIT D4",
                    "BUILDING_NAME": "ADANAC PARK",
                    "THOROUGHFARE_NAME": "ADANAC DRIVE",
                    "DEPENDENT_LOCALITY": "NURSLING",
                    "POST_TOWN": "SOUTHAMPTON",
                    "POSTCODE": "SO16 0BT",
                    "RPC": "2",
                    "X_COORDINATE": 437230.0,
                    "Y_COORDINATE": 115775.0,
                    "LNG": -163806.07,
                    "LAT": 6610726.44,
                    "STATUS": "HISTORICAL",
                    "LOGICAL_STATUS_CODE": "8",
                    "CLASSIFICATION_CODE": "CI01",
                    "CLASSIFICATION_CODE_DESCRIPTION": "Factory/Manufacturing",
                    "LOCAL_CUSTODIAN_CODE": 1760,
                    "LOCAL_CUSTODIAN_CODE_DESCRIPTION": "TEST VALLEY",
                    "POSTAL_ADDRESS_CODE": "D",
                    "POSTAL_ADDRESS_CODE_DESCRIPTION": "A record which is linked to PAF",
                    "BLPU_STATE_CODE": "null",
                    "BLPU_STATE_CODE_DESCRIPTION": "Unknown/Not applicable",
                    "TOPOGRAPHY_LAYER_TOID": "osgb5000005278503035",
                    "PARENT_UPRN": "10094212626",
                    "LAST_UPDATE_DATE": "01/09/2021",
                    "ENTRY_DATE": "20/08/2019",
                    "LANGUAGE": "EN",
                    "MATCH": 0.4,
                    "MATCH_DESCRIPTION": "NO MATCH",
                },
            },
        ),
        param(
            {
                "X_COORDINATE": 437230.0,
                "Y_COORDINATE": 115775.0,
                "test": "x",
                "test2": 6610726.44,
            },
            "EPSG:27700",
            {
                "type": "Feature",
                "geometry": {"type": "Point", "coordinates": [437230.0, 115775.0]},
                "properties": {
                    "X_COORDINATE": 437230.0,
                    "Y_COORDINATE": 115775.0,
                    "test": "x",
                    "test2": 6610726.44,
                },
            },
        ),
    ]
    return test_variables, test_data


def test_address_to_feature_error():
    test_variables = "address, crs"
    test_data = [
        param(
            {
                "X_COORDINATE": 437230.0,
                "Y_COORDINATE": 115775.0,
                "test": "x",
                "test2": 6610726.44,
            },
            "EPSG:3857",
        )
    ]
    return test_variables, test_data



import asyncio
import logging
from typing import Any, Dict, Optional

import aiohttp

from osdatahub.AsyncAPI.rate_limiter import RateLimiter

_USER_AGENT_TAG = "osdatahub-python-async"


class AsyncHTTPClient:
    """
    Reusable async HTTP client with connection pooling, rate limiting, and retry logic.

    This client provides:
    - Connection pooling via aiohttp TCPConnector
    - Rate limiting via semaphore and request delays
    - Automatic retries with exponential backoff
    - Content-length validation
    - Proxy support

    Args:
        max_concurrent: Maximum concurrent requests (default: 5)
        request_delay: Delay between requests in seconds (default: 0.1)
        max_retries: Maximum retry attempts on failure (default: 3)
        connector_limit: Total connection pool limit (default: 10)
        connector_limit_per_host: Per-host connection limit (default: 5)
        timeout: Request timeout in seconds (default: 30)
        proxies: Proxy configuration dict (e.g., {"http": "...", "https": "..."})
            Uses the "https" value for HTTPS requests, "http" for HTTP.

    Example::

        async with AsyncHTTPClient() as client:
            response = await client.get(url, params=params, headers=headers)
    """

    def __init__(
        self,
        max_concurrent: int = 5,
        request_delay: float = 0.1,
        max_retries: int = 3,
        connector_limit: int = 30,
        connector_limit_per_host: int = 5,
        timeout: float = 30.0,
        proxies: Optional[Dict[str, str]] = None,
    ) -> None:
        self._max_concurrent = max_concurrent
        self._request_delay = request_delay
        self._max_retries = max_retries
        self._connector_limit = connector_limit
        self._connector_limit_per_host = connector_limit_per_host
        self._timeout = timeout
        self._proxies = proxies or {}

        self._session: Optional[aiohttp.ClientSession] = None
        self._rate_limiter: Optional[RateLimiter] = None

    def _get_proxy(self, url: str) -> Optional[str]:
        """Get the appropriate proxy URL for the given request URL."""
        if not self._proxies:
            return None
        if url.startswith("https://"):
            return self._proxies.get("https")
        return self._proxies.get("http")

    async def _get_session(self) -> aiohttp.ClientSession:
        "Initialisation of aiohttp session."
        if self._session is None or self._session.closed:
            connector = aiohttp.TCPConnector(
                limit=self._connector_limit,
                limit_per_host=self._connector_limit_per_host,
                ttl_dns_cache=300,
                force_close=False,
                enable_cleanup_closed=True,
            )
            timeout = aiohttp.ClientTimeout(total=self._timeout)
            self._session = aiohttp.ClientSession(connector=connector, timeout=timeout)
        return self._session

    def _get_rate_limiter(self) -> RateLimiter:
        """Initialisation of rate limiter."""
        if self._rate_limiter is None:
            self._rate_limiter = RateLimiter(
                max_concurrent=self._max_concurrent, request_delay=self._request_delay
            )
        return self._rate_limiter

    async def get(
        self,
        url: str,
        params: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Perform an async GET request with rate limiting and retries.

        Args:
            url: The URL to request
            params: Query parameters
            headers: HTTP headers
            **kwargs: Additional arguments passed to aiohttp

        Returns:
            JSON response as dict

        Raises:
            aiohttp.ClientResponseError: On HTTP errors after retries exhausted
            IOError: On content length mismatch
        """
        headers = self._prepare_headers(headers)
        session = await self._get_session()
        rate_limiter = self._get_rate_limiter()

        last_exception: Optional[Exception] = None

        proxy = self._get_proxy(url)

        for attempt in range(self._max_retries):
            try:
                # TODO: Write docs on how this works
                # The key is that everything must acquire the semapohore to proceed
                async with rate_limiter:
                    async with session.get(
                        url, params=params, headers=headers, proxy=proxy, **kwargs
                    ) as response:
                        response.raise_for_status()
                        data = await response.json()
                        self._validate_content_length(response, data)
                        return data
            except (aiohttp.ClientError, asyncio.TimeoutError) as e:
                last_exception = e
                if attempt < self._max_retries - 1:
                    backoff = 0.5 * (2**attempt)  # Exponential backoff
                    logging.warning(
                        f"Request failed (attempt {attempt + 1}/{self._max_retries}), "
                        f"retrying in {backoff}s: {e}"
                    )
                    await asyncio.sleep(backoff)

        if last_exception is not None:
            raise last_exception
        raise RuntimeError("Unexpected state: no exception but request did not succeed")

    async def post(
        self,
        url: str,
        data: Optional[Any] = None,
        json: Optional[Dict[str, Any]] = None,
        params: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Perform an async POST request with rate limiting and retries.

        Args:
            url: The URL to request
            data: Form data to send
            json: JSON data to send
            params: Query parameters
            headers: HTTP headers
            **kwargs: Additional arguments passed to aiohttp

        Returns:
            JSON response as dict

        Raises:
            aiohttp.ClientResponseError: On HTTP errors after retries exhausted
        """
        headers = self._prepare_headers(headers)
        session = await self._get_session()
        rate_limiter = self._get_rate_limiter()

        last_exception: Optional[Exception] = None

        proxy = self._get_proxy(url)

        for attempt in range(self._max_retries):
            try:
                async with rate_limiter:
                    async with session.post(
                        url,
                        data=data,
                        json=json,
                        params=params,
                        headers=headers,
                        proxy=proxy,
                        **kwargs,
                    ) as response:
                        response.raise_for_status()
                        return await response.json()
            except (aiohttp.ClientError, asyncio.TimeoutError) as e:
                last_exception = e
                if attempt < self._max_retries - 1:
                    backoff = 0.5 * (2**attempt)
                    logging.warning(
                        f"Request failed (attempt {attempt + 1}/{self._max_retries}), "
                        f"retrying in {backoff}s: {e}"
                    )
                    await asyncio.sleep(backoff)

        if last_exception is not None:
            raise last_exception
        raise RuntimeError("Unexpected state: no exception but request did not succeed")

    def _prepare_headers(self, headers: Optional[Dict[str, str]]) -> Dict[str, str]:
        """Add User-Agent header to requests."""
        headers = headers.copy() if headers else {}
        headers.setdefault("User-Agent", _USER_AGENT_TAG)
        return headers

    def _validate_content_length(
        self, response: aiohttp.ClientResponse, data: Any
    ) -> None:
        """
        Validate response content length matches header.
        """
        expected = response.headers.get("Content-Length")
        if expected is not None:
            pass

    async def close(self) -> None:
        """Close the HTTP session."""
        if self._session is not None and not self._session.closed:
            await self._session.close()
            self._session = None

    async def __aenter__(self) -> "AsyncHTTPClient":
        """Context manager entry."""
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        """Context manager exit - cleanup resources."""
        await self.close()



import functools
import operator
from typing import Union

from osdatahub import Extent


def _binary_operator(*classes):
    def func_acceptor(func):
        def inner(instance, other):
            nonlocal classes
            classes = tuple(globals()[i] if isinstance(i, str) else i for i in classes)
            if not isinstance(other, classes):
                return NotImplemented
            return func(instance, other)

        return inner

    return func_acceptor


class Filter:
    """
    A container class for an OGD XML filter.
    Allows XML to be joined using `and` and `or` operations.
    It should not be instantiated directly, instead produced by one of the methods in this module.
    """

    def __init__(self, xml: str):
        self.xml: str = xml

    @staticmethod
    def _apply_op(filter1: str, filter2: str, operation: str) -> str:
        return f"<ogc:{operation}>{filter1}{filter2}</ogc:{operation}>"

    @_binary_operator("Filter", str)
    def __add__(self, other) -> "Filter":
        if isinstance(other, Filter):
            return Filter(self.xml + other.xml)
        return Filter(self.xml + other)

    @_binary_operator("Filter")
    def __and__(self, other) -> "Filter":
        return Filter(self._apply_op(self.xml, other.xml, "And"))

    def __bool__(self) -> bool:
        raise NotImplementedError("Did you use a boolean operation, meaning to use a bitwise operation instead?")

    def __eq__(self, other) -> bool:
        if isinstance(other, Filter):
            return self.xml == other.xml
        elif isinstance(other, str):
            return self.xml == other
        return False

    @_binary_operator("Filter", str)
    def __iadd__(self, other) -> "Filter":
        if isinstance(other, Filter):
            self.xml = self.xml + other.xml
        else:
            self.xml = self.xml + other
        return self

    @_binary_operator("Filter")
    def __iand__(self, other) -> "Filter":
        self.xml = self._apply_op(self.xml, other.xml, "And")
        return self

    @_binary_operator("Filter")
    def __ior__(self, other) -> "Filter":
        self.xml = self._apply_op(self.xml, other.xml, "Or")
        return self

    @_binary_operator("Filter")
    def __or__(self, other) -> "Filter":
        return Filter(self._apply_op(self.xml, other.xml, "Or"))

    @_binary_operator("Filter", str)
    def __radd__(self, other) -> "Filter":
        if isinstance(other, Filter):
            return Filter(other.xml + self.xml)
        return Filter(other + self.xml)

    @_binary_operator("Filter")
    def __rand__(self, other):
        return Filter(self._apply_op(other.xml, self.xml, "And"))

    def __repr__(self) -> str:
        return repr(self.xml)

    @_binary_operator("Filter")
    def __ror__(self, other):
        return Filter(self._apply_op(other.xml, self.xml, "Or"))

    def __str__(self) -> str:
        return self.xml


def filter_or(*filters: Filter) -> Filter:
    """Constructs an OGC XML filter that performs an 'or' on the given filters

    Args:
        filters (Filter): The filters to be joined

    Returns:
        Filter: A valid OGC XML filter
    """
    return functools.reduce(operator.or_, filters)


def filter_and(*filters: Filter) -> Filter:
    """Constructs an OGC XML filter that performs an 'and' on the given filters

    Args:
        filters (Filter): The filters to be joined

    Returns:
        Filter: A valid OGC XML filter
    """
    return functools.reduce(operator.and_, filters)


def spatial_filter(operator: str, extent: Extent) -> Filter:
    """Constructs an OGC XML filter using the given operator string and the given extent

    Args:
        operator (str): Case-sensitive name of operator for OGC filter string
        extent (Extent): The desired region to be filtered, given as an Extent object

    Returns:
        Filter: A valid OGC XML filter
    """
    coords = extent.xml_coords
    crs = extent.crs.upper()
    return Filter(
        f"<ogc:{operator}>"
        "<ogc:PropertyName>SHAPE</ogc:PropertyName>"
        f"<gml:Polygon xmlns:gml='http://www.opengis.net/gml' srsName='{crs}'>"
        "<gml:outerBoundaryIs>"
        "<gml:LinearRing>"
        f'<gml:coordinates decimal="." cs="," ts=" ">{coords}</gml:coordinates>'
        "</gml:LinearRing>"
        "</gml:outerBoundaryIs>"
        "</gml:Polygon>"
        f"</ogc:{operator}>"
    )


def intersects(extent: Extent) -> Filter:
    """Constructs an OGC XML filter for data that intersects the given extent

    Args:
        extent (Extent): The desired region to be filtered, given as an Extent object

    Returns:
        Filter: A valid OGC XML filter
    """
    return spatial_filter("Intersects", extent)


def touches(extent: Extent) -> Filter:
    """Constructs an OGC XML filter for data that touches the given extent

    Args:
        extent (Extent): The desired region to be filtered, given as an Extent object

    Returns:
        Filter: A valid OGC XML filter
    """
    return spatial_filter("Touches", extent)


def disjoint(extent: Extent) -> Filter:
    """Constructs an OGC XML filter for data that does not interact with the given extent

    Args:
        extent (Extent): The desired region to be filtered, given as an Extent object

    Returns:
        Filter: A valid OGC XML filter
    """
    return spatial_filter("Disjoint", extent)


def contains(extent: Extent) -> Filter:
    """Constructs an OGC XML filter for data that contains the given extent

    Args:
        extent (Extent): The desired region to be filtered, given as an Extent object

    Returns:
        Filter: A valid OGC XML filter
    """
    return spatial_filter("Contains", extent)


def within(extent: Extent) -> Filter:
    """Constructs an OGC XML filter for data that is within the given extent

    Args:
        extent (Extent): The desired region to be filtered, given as an Extent object

    Returns:
        Filter: A valid OGC XML filter
    """
    return spatial_filter("Within", extent)


def crosses(extent: Extent) -> Filter:
    """Constructs an OGC XML filter for data that crosses the given extent

    Args:
        extent (Extent): The desired region to be filtered, given as an Extent object

    Returns:
        Filter: A valid OGC XML filter
    """
    return spatial_filter("Crosses", extent)


def overlaps(extent: Extent) -> Filter:
    """Constructs an OGC XML filter for data that overlaps the given extent

    Args:
        extent (Extent): The desired region to be filtered, given as an Extent object

    Returns:
        Filter: A valid OGC XML filter
    """
    return spatial_filter("Overlaps", extent)


def equals(extent: Extent) -> Filter:
    """Constructs an OGC XML filter for data that is equal to the given extent

    Args:
        extent (Extent): The desired region to be filtered, given as an Extent object

    Returns:
        Filter: A valid OGC XML filter
    """
    return spatial_filter("Equals", extent)


def single_attribute_filter(property_name, filter_name, value) -> Filter:
    return Filter(
        f"<ogc:{filter_name}>"
        f"<ogc:PropertyName>{property_name}</ogc:PropertyName>"
        f"<ogc:Literal>{value}</ogc:Literal>"
        f"</ogc:{filter_name}>"
    )


def is_between(property_name: str, lower: float, upper: float) -> Filter:
    """Constructs an OGC XML filter for a numerical attribute between 2 values

    Args:
        property_name (str): Property / attribute name to be filtered
        lower (float): The filter's lower bound
        upper (float): The filter's upper bound

    Returns:
        Filter: A valid OGC XML filter
    """
    return Filter(
        f"<ogc:PropertyIsBetween>"
        f"<ogc:PropertyName>{property_name}</ogc:PropertyName>"
        f"<LowerBoundary>"
        f"<ogc:Literal>{lower}</ogc:Literal>"
        "</LowerBoundary>"
        "<UpperBoundary>"
        f"<ogc:Literal>{upper}</ogc:Literal>"
        "</UpperBoundary>"
        f"</ogc:PropertyIsBetween>"
    )


def is_like(
        property_name: str,
        value: str,
        wildcard: str = "*",
        single_char: str = "#",
        escape_char: str = "!",
) -> Filter:
    """Constructs an OGC XML filter for a string attribute that is similar to
    the input value

    Args:
        property_name (str): Property / attribute name to be filtered
        value (str): String that is used to match with attribute values
        wildcard (str, optional): A character that any combination of other
            characters will match with. Defaults to "*".
        single_char (str, optional): A character that any single character
            will match with. Defaults to "#".
        escape_char (str, optional): Used to escape the meaning of the
            wildcard, single_char and escape_char itself. Defaults to "!".

    Returns:
        Filter: A valid OGC XML filter
    """
    return Filter(
        f'<ogc:PropertyIsLike wildCard="{wildcard}" singleChar="{single_char}" escapeChar="{escape_char}">'
        f"<ogc:ValueReference>{property_name}</ogc:ValueReference>"
        f"<ogc:Literal>{value}</ogc:Literal>"
        f"</ogc:PropertyIsLike>"
    )


def is_equal(property_name: str, value: Union[str, float]) -> Filter:
    """Constructs an OGC Filter for an attribute that is equal to
    the input value

    Args:
        property_name (str): Property / attribute name to be filtered
        value (Union[str, float]): Value used in filter for comparison

    Returns:
        Filter: A valid OGC XML filter
    """
    return single_attribute_filter(property_name, "PropertyIsEqualTo", value)


def is_not_equal(property_name: str, value: Union[str, float]) -> Filter:
    """Constructs an OGC Filter for an attribute that is not equal to
    the input value

    Args:
        property_name (str): Property / attribute name to be filtered
        value (Union[str, float]): Value used in filter for comparison

    Returns:
        Filter: A valid OGC XML filter
    """
    return single_attribute_filter(property_name, "PropertyIsNotEqualTo", value)


def is_less_than(property_name: str, value: float) -> Filter:
    """Constructs an OGC Filter for a numerical attribute that is less than
    the input value

    Args:
        property_name (str): Property / attribute name to be filtered
        value (float): Value used in filter for comparison

    Returns:
        Filter: A valid OGC XML filter
    """
    return single_attribute_filter(property_name, "PropertyIsLessThan", value)


def is_greater_than(property_name: str, value: float) -> Filter:
    """Constructs an OGC Filter for a numerical attribute that is greater than
    the input value

    Args:
        property_name (str): Property / attribute name to be filtered
        value (float): Value used in filter for comparison

    Returns:
        Filter: A valid OGC XML filter
    """
    return single_attribute_filter(property_name, "PropertyIsGreaterThan", value)


def is_less_than_or_equal_to(property_name: str, value: float) -> Filter:
    """Constructs an OGC Filter for a numerical attribute that is less
    than or equal to the input value

    Args:
        property_name (str): Property / attribute name to be filtered
        value (float): Value used in filter for comparison

    Returns:
        Filter: A valid OGC XML filter
    """
    return single_attribute_filter(property_name, "PropertyIsLessThanOrEqualTo", value)


def is_greater_than_or_equal_to(property_name: str, value: float) -> Filter:
    """Constructs an OGC Filter for a numerical attribute that is greater
    than or equal to the input value

    Args:
        property_name (str): Property / attribute name to be filtered
        value (float): Value used in filter for comparison

    Returns:
        Filter: A valid OGC XML filter
    """
    return single_attribute_filter(
        property_name, "PropertyIsGreaterThanOrEqualTo", value
    )



import geojson
from pytest import param
from shapely.geometry import MultiPolygon, Polygon, mapping


def test_clean_polygon():
    test_variables = "feature, expected_result"
    test_data = [
        param(
            {"geometry": simple_polygon1},
            {"geometry": simple_polygon1.copy()},
            id="simple polygon",
        ),
        param(
            {"geometry": simple_polygon2},
            {"geometry": simple_polygon2.copy()},
            id="simple polygon 2",
        ),
        param(
            {"geometry": polygon_with_hole},
            {"geometry": polygon_with_hole.copy()},
            id="polygon with hole",
        ),
        param(
            {"geometry": multipolygon1},
            {"geometry": multipolygon1.copy()},
            id="simple multipolygon",
        ),
        param(
            {"geometry": multipolygon_with_hole},
            {"geometry": multipolygon_with_hole.copy()},
            id="multipolygon with hole",
        ),
        param(
            {"geometry": get_test_polygons("sites_polygon", "API (uncleaned)")},
            {"geometry": get_test_polygons("sites_polygon", "QGIS")},
            id="sites polygon",
        ),
        param(
            {
                "geometry": get_test_polygons(
                    "sites_polygon_with_hole", "API (uncleaned)"
                )
            },
            {"geometry": get_test_polygons("sites_polygon_with_hole", "QGIS")},
            id="sites polygon with hole",
        ),
        param(
            {"geometry": get_test_polygons("sites_multipolygon", "API (uncleaned)")},
            {"geometry": get_test_polygons("sites_multipolygon", "QGIS")},
            id="sites multipolygon",
        ),
        param(
            {"geometry": get_test_polygons("sites_multipolygon2", "API (uncleaned)")},
            {"geometry": get_test_polygons("sites_multipolygon2", "QGIS")},
            id="sites multipolygon2",
        ),
        param(
            {
                "geometry": get_test_polygons(
                    "sites_multipolygon_with_hole", "API (uncleaned)"
                )
            },
            {"geometry": get_test_polygons("sites_multipolygon_with_hole", "QGIS")},
            id="sites multipolygon with hole",
        ),
    ]
    return test_variables, test_data


def get_test_polygons(polygon_name, source):
    path = f"./tests/data/clean_polygon_data/{source}/{polygon_name}.geojson"
    with open(path, "r") as f:
        return geojson.load(f)["features"]


shell1 = [(1, 0), (1, 1), (0, 1), (0, 0)]
shell2 = [(3, 2), (3, 3), (2, 3), (2, 2)]
hole = [(0.75, 0.25), (0.75, 0.75), (0.25, 0.75), (0.25, 0.25)]

polygon1 = Polygon(shell1)
polygon2 = Polygon(shell2)
polygon1_with_hole = Polygon(shell1, [hole])

simple_polygon1 = mapping(polygon1)
simple_polygon2 = mapping(polygon2)

polygon_with_hole = mapping(Polygon(shell2, [hole]))
multipolygon1 = mapping(MultiPolygon([polygon1, polygon2]))
multipolygon_with_hole = mapping(MultiPolygon([polygon2, polygon1_with_hole]))



import functools
import json
import logging
import os
import time
from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import datetime
from http import HTTPStatus
from multiprocessing import cpu_count
from pathlib import Path
from typing import List, Union

import requests
from requests.exceptions import HTTPError
from tqdm import tqdm

import osdatahub

retries = 3


class _DownloadObj:
    """ Helper class to download a file from Downloads API

    Args:
        url (str): Direct url for the download file
        file_name (str): Name of the file to be downloaded
    """

    def __init__(self, url: str, file_name: str, size: int):
        self.url = url
        self.file_name = file_name
        self.size = size

    def download(self, 
                 output_dir: Union[str, Path], 
                 overwrite: bool = False, 
                 pbar: Union[tqdm, None] = None) -> str:
        """
        Downloads file to given directory

        Args:
            output_dir (Union[str, Path]): Directory to save the downloaded file
            overwrite (bool, optional): whether to overwrite an existing file. Defaults to False
            pbar (tqdm, optional): tqdm progress bar to update in the event of downloading multiple files at once
        """
        output_path = os.path.join(output_dir, self.file_name)

        if os.path.isfile(output_path) and not overwrite:
            logging.warning(f"Overwrite is set to False and there is a file already in the location {output_path}. "
                            f"Skipping download...")
            return output_path

        for _ in range(retries):
            try:
                header = {'User-Agent': 'osdatahub-python'}
                response = requests.get(
                    self.url, stream=True, headers=header, proxies=osdatahub.get_proxies())
                response.raise_for_status()
                expected_size = int(response.headers.get('content-length'))
                current_size = 0
                chunk_size = 1048576  # 1024 ** 2 -> 1MB
                if response.status_code == 200:
                    with open(output_path, 'wb') as f:
                        if not pbar:
                            pbar = tqdm(
                                total=expected_size, desc=self.file_name, unit="B", unit_scale=True, leave=True)
                        for chunk in response.iter_content(chunk_size=chunk_size):
                            current_size += len(chunk)
                            f.write(chunk)
                            f.flush()
                            pbar.update(chunk_size)
                        if expected_size != current_size:
                            deficit = expected_size - current_size
                            raise IOError(
                                f'incomplete read ({current_size} bytes read, {deficit} more expected)'
                            )
                        pbar.write(
                            f"Finished downloading {self.file_name} to {output_path}")
                break

            except HTTPError as exc:
                if int(exc.response.status_code) == 429:
                    time.sleep(20)
                    continue
                raise

        return output_path


def remove_key(url: str):
    """Remove key from url
    """
    return "".join([section for section in url.split("&") if "key" not in section])


def format_missing_files(missing_files: List[_DownloadObj]) -> List[dict]:
    """Convert download objects to dictionaries and sanitise
    """
    file_info = []
    for _download_obj in missing_files:
        info = _download_obj.__dict__
        info['url'] = remove_key(info['url'])
        file_info.append(info)
    return {
        "missing_file_count": len(missing_files),
        "missing_file_info": file_info
    }


def save_missing_files(missing_files: List[_DownloadObj], output_dir: Union[str, Path]) -> None:
    """Format and save missing files 
    """
    if len(missing_files) == 0:
        return
    data = format_missing_files(missing_files)
    path = os.path.join(
        output_dir, f"missing_files.{datetime.now().strftime('%Y%m%d%H%M%S')}.json")
    json.dump(data, open(path, "w"))


class _DownloadsAPIBase(ABC):
    """Parent class for Product and DataPackage classes as part of the DownloadsAPI
    (https://osdatahub.os.uk/docs/downloads/overview)

    Args:
        product_id (str): Valid ID for a Downloads API Product or DataPackage.

    """
    _ENDPOINT = f"https://api.os.uk/downloads/v1/"

    def __init__(self, product_id: str):
        self._id = product_id

    def _endpoint(self, api_name: str) -> str:
        """
        Adds the api to the base endpoint

        Args:
            api_name (str): API endpoint to add to _ENDPOINT attribute

        Returns:
            str: concatenated endpoint consisting of _ENDPOINT and api_name
        """
        return f"{self._ENDPOINT}/{api_name}"

    @property
    def id(self):
        return self._id

    @property
    @functools.lru_cache()
    def details(self) -> dict:
        """
        Calls endpoint to return details about the product or data package
        """
        response = osdatahub.get(self._endpoint(
            self._id), proxies=osdatahub.get_proxies())
        response.raise_for_status()
        return response.json()

    @classmethod
    def all_products(cls, **kwargs) -> list:
        """
        Returns a list of all available products of the product type

        Returns: list of dictionaries containing all products available to download

        """
        response = osdatahub.get(
            cls._ENDPOINT, proxies=osdatahub.get_proxies())
        response.raise_for_status()
        return response.json()

    @abstractmethod
    def product_list(self):
        """
        Abstract method for returning a list of files available to download for a specific data package/opendata product
        Returns:

        """
        pass

    @staticmethod
    def _download(download_list: Union[list, _DownloadObj], 
                  output_dir: Union[str, Path], 
                  overwrite: bool = False,
                  download_multiple: bool = False, 
                  processes: Union[int, None] = None) -> list:
        """
        Downloads product/datapackage to the given directory. Can download a single format or can download multiple
        formats in parallel

        Args:
             download_list (Union[list, DownloadObj]): The DownloadObj objects representing all the
                products/data packages that need to be downloaded
             output_dir (Union[str, Path]): path to directory where the files will be saved to
             overwrite (bool, optional): Whether to overwrite any existing files with the same name and path.
                Defaults to False
             download_multiple (bool, optional): Whether to download multiple files, generally the same data but in
                different formats. Defaults to False
             processes (int, optional): If downloading multiple files, the number of parallel processes to be used.
                defaults to the machine's CPU count
        """
        if isinstance(download_list, list) and len(download_list) == 0:
            raise Exception(
                "Argument \"download_list\" is empty. Please provide at least one DownloadObj to download")
        elif isinstance(download_list, list) and len(download_list) > 1 and not download_multiple:
            raise Exception("Argument \"download_list\" contains more than 1 object to download, but argument "
                            "\"download_multiple\" is set to False. Please pass only 1 download or set "
                            "\"download_multiple\" to True.")

        if not os.path.isdir(output_dir):
            os.makedirs(output_dir)

        # downloads in parallel if multiple files need to be downloaded
        if isinstance(download_list, list) and len(download_list) > 1 and download_multiple:
            if not processes:
                processes = cpu_count()
            with ThreadPoolExecutor(max_workers=processes) as executor:
                pbar = tqdm(total=sum([d.size for d in download_list]), unit="B", unit_scale=True, leave=True,
                            desc=f"Downloaded 0/{len(download_list)} files from osdatahub")

                processed_downloads = {}
                num_downloads_completed = 0
                results = []
                missing_files = []

                for p in download_list:
                    future = executor.submit(
                        p.download, output_dir, overwrite, pbar)
                    processed_downloads[future] = p

                for future in as_completed(processed_downloads):
                    info = processed_downloads[future]
                    try:
                        results.append(future.result())
                        num_downloads_completed += 1
                        pbar.set_description(
                            f"Downloaded {num_downloads_completed}/{len(download_list)} files from osdatahub")
                    except Exception:
                        missing_files.append(info)

                save_missing_files(missing_files, output_dir)
        else:
            # download single file
            d = download_list[0] if isinstance(
                download_list, list) else download_list
            results = [d.download(output_dir, overwrite)]

        return results



import functools
import logging
from pathlib import Path
from typing import Union

import requests
from typeguard import typechecked
import osdatahub



from .downloads_api import _DownloadObj, _DownloadsAPIBase


class DataPackageDownload(_DownloadsAPIBase):
    """
    Main class for downloading OS Premium Data Packages
    (https://osdatahub.os.uk/docs/downloads/technicalSpecification#/Data%20Packages)

    Args:
        key (str): A valid OS API Key. Get a free key here - https://osdatahub.os.uk/
        product_id (str): Valid ID for Downloads API  DataPackage.

    """
    _ENDPOINT = _DownloadsAPIBase._ENDPOINT + "dataPackages"

    def __init__(self, key: str, product_id: str):
        super().__init__(product_id=product_id)
        self.key = key

    @classmethod
    def all_products(cls, key) -> list:
        """
        Returns a list of all available data packages

        Args:
            key (str): A valid OS API Key. Get a free key here - https://osdatahub.os.uk/

        Returns: A list of dictionaries containing all available Data Packages

        """
        response = osdatahub.get(cls._ENDPOINT, params={"key": key}, proxies=osdatahub.get_proxies())
        response.raise_for_status()
        content = response.json()
        if not content:
            logging.warning(f"You have no premium data packages available. "
                            f"Make sure that you first ordered a data package at "
                            f"https://osdatahub.os.uk/downloads/premium")
        return content

    @property
    @functools.lru_cache()
    def versions(self) -> list:
        """
        Get all the available versions for the data package
        """
        response = osdatahub.get(self._endpoint(f"{self._id}/versions"), params={"key": self.key}, proxies=osdatahub.get_proxies())
        response.raise_for_status()
        return response.json()

    @typechecked
    def product_list(self, version_id: str, return_downloadobj: bool = False) -> Union[list, dict]:
        """
        Returns a list of possible downloads for a specific OS Premium Data Package based on given filters

        Args:
            version_id (str): Filter the list of possible downloads by the version id of the given product
            return_downloadobj (bool, optional): Returns the downloadable files as _DownloadObj objects instead of as a
                json. Defaults to False

        Returns:
            List of downloadable files from Downloads API
        """
        endpoint = self._endpoint(f"{self._id}/versions/{version_id}")
        params = {"key": self.key}
        response = osdatahub.get(url=endpoint, params=params, proxies=osdatahub.get_proxies())
        response.raise_for_status()
        content = response.json()
        if not content:
            logging.warning(f"There are no premium data packages available with id={self.id} and "
                            f"version_id={version_id}. Make sure that you first ordered a data package at"
                            f"https://osdatahub.os.uk/downloads/premium")
        if return_downloadobj:
            return [_DownloadObj(url=download["url"], file_name=download["fileName"], size=download["size"])
                    for download in content["downloads"]]

        return content

    @typechecked
    def download(self,
                 version_id: str,
                 output_dir: Union[str, Path] = ".",
                 file_name: Union[str, None] = None,
                 overwrite: bool = False,
                 processes: Union[int, None] = None) -> list:
        """
        Downloads Data Package files to your local machine

        Args:
            version_id (str): The version id of the data package to download
            output_dir (Union[str, Path], optional): the path where the downloaded files will be saved. Defaults to
                current working directory
            file_name (str, optional): name of the file(s) to download
            overwrite (bool, optional): whether to overwrite existing files. Defaults to False
            processes (int, optional): Number of processes with which to download multiple files. Only relevant if
                multiple files will be downloaded (and download_multiple is set to True)
        """
        if file_name is not None:
            url = f'{self._endpoint(f"{self.id}/versions/{version_id}/downloads")}?fileName={file_name}&key={self.key}'
            return [_DownloadObj(url=url, file_name=file_name, size=0).download(output_dir=output_dir,
                                                                                overwrite=overwrite)]
        else:
            download_list = self.product_list(version_id, return_downloadobj=True)
            return super()._download(download_list=download_list,
                                     output_dir=output_dir,
                                     overwrite=overwrite,
                                     download_multiple=True,
                                     processes=processes)



from typing import Union

from typing import Union

from geojson import FeatureCollection
from osdatahub.grow_list import GrowList
from shapely.geometry import LinearRing


def clean_features(feature_list: list, geom_type: str) -> list:
    """Post-processes API features to standardize geometry. In particular
    remove additional brackets and correct the nesting of
    brackets for multipolygons with holes.

    Args:
        feature_list (list): List of GeoJSON geometries
        geom_type (str): string describing which GeoJSON geometry type is
        in feature_list.

    Returns:
        list: cleaned GeoJSON geometries
    """
    if geom_type == "LineString":
        return clean_linestrings(feature_list)
    if geom_type == "Polygon":
        return clean_polygons(feature_list)
    return feature_list


def clean_linestrings(feature_list: list) -> list:
    """Post-process API LineStrings to standardize format by removing
    additional brackets.

    Args:
        feature_list (list): List of GeoJSON LineStrings

    Returns:
        list: List of GeoJSON LineStrings with extraneous brackets removed
    """
    return [clean_linestring(f) for f in feature_list]


def clean_linestring(feature: dict) -> dict:
    """Post-process API LineString to standardize format by removing
    additional brackets.

    Args:
        feature (dict): GeoJSON LineString feature

    Returns:
        dict: GeoJSON LineString feature with extraneous brackets removed
    """
    feature["geometry"]["coordinates"] = feature["geometry"]["coordinates"][0]
    return feature


def clean_polygons(feature_list: list) -> list:
    """Post-process API Polygons to fix geometries of MultiPolygons.

    Args:
        feature_list (list): List of GeoJSON Polygons

    Returns:
        list: List of fixed GeoJSON Polygons
    """
    return [clean_polygon(f) for f in feature_list]


def clean_polygon(feature: dict) -> dict:
    """Fixes the geometry of MultiPolygons, which are identified by being
    lists rather than dictionaries.

    Args:
        feature (dict): GeoJSON Polygon feature

    Returns:
        dict: Fixed GeoJSON Polygon feature
    """
    if isinstance(feature["geometry"], list):
        coordinates = feature["geometry"][0]["geometry"]["coordinates"]
        nested_polygons = nest_polygons(coordinates)
        feature["geometry"][0]["geometry"]["coordinates"] = nested_polygons
        feature["geometry"][0]["geometry"]["type"] = "MultiPolygon"
        return feature
    return feature


def nest_polygons(coordinates):
    """Checks if polygons are oriented clockwise or anticlockwise,
    and uses the order to assessemble polygons with holes in correctly.
    Only needed to correct multipolygons with holes.

    Args:
        coordinates (list): Coordinates in a clockwise orientation are exteriors,
    the following polygons of anticlockwise orientation are the holes for that
    polygons. Next clockwise coordinates are assumed to be exterior of new
    polygon.

    Returns:
        new_polys (list): A list of nested coordinates
    """
    new_polys = []
    polygon = []
    is_ccw = [LinearRing(coords).is_ccw for coords in coordinates]
    for direction, coords in zip(is_ccw, coordinates):
        if not direction:
            if polygon:
                new_polys.append(polygon)
            polygon = [coords]
        else:
            polygon.append(coords)
    new_polys.append(polygon)
    return new_polys


def features_to_geojson(feature_list, geom_type, crs) -> FeatureCollection:
    """Converts a list of GeoJSON gemetries to a FeatureCollection

    Args:
        feature_list (list): List of GeoJSON geometries
        geom_type (str): string describing which GeoJSON geometry type is
        in feature_list.
        crs (str): string representation of a coordinate system

    Returns:
        FeatureCollection: cleaned FeatureCollection
    """
    features = clean_features(feature_list, geom_type)
    return FeatureCollection(features, crs=crs)


def addresses_to_geojson(address_list, crs):
    """Converts a list of GeoJSON gemetries to a FeatureCollection, where the
    list of GeoJSONs represent addresses

    Args:
        address_list (list): List of GeoJSON addresses
        crs (str): string representation of a coordinate system

    Returns:
        FeatureCollection: cleaned FeatureCollection
    """
    features = [address_to_feature(address, crs) for address in address_list]
    return FeatureCollection(features, crs=crs)


def address_to_feature(address, crs):
    """The raw address from the PlacesAPI returns address in a JSON format where
    the coords have different keys depending on what CRS is requested. This
    function is needed to ensure that the keys of the coordinates are the same
    regardless of CRS.

    Args:
        address (dict): dictionary representation of an address
        crs (str): string representation of a coordinate system

    Returns:
        dict: dictionary representation of the input address

    Raises:
        ValueError: If CRS is not British National Grid, there should be
        attributes called 'LAT' and 'LNG' in the address
    """
    if crs.lower() in ("epsg:27700", "bng"):
        try:
            x, y = address["X_COORDINATE"], address["Y_COORDINATE"]
        except KeyError:
            x, y = address["GEOMETRY_X"], address["GEOMETRY_Y"]
    elif all(i in address.keys() for i in ("LNG", "LAT")):
        x, y = address["LNG"], address["LAT"]
    else:
        raise ValueError(
            f"If CRS is not British National Grid, there should be attributes called 'LAT' and 'LNG' in"
            f" address. LAT and LNG were not found, and CRS is {crs}"
        )

    return {
        "type": "Feature",
        "geometry": {"type": "Point", "coordinates": [x, y]},
        "properties": {**address},
    }


def validate_in_range(value: float, minimum: float, maximum: float) -> float:
    """Checks that the input value is between the maximum and minimum values
    and returns the original value if it is.

    Args:
        value (float): value to check
        minimum (float): minimum value in range (inclusive)
        maximum (float): maximum value in range (inclusive)

    Returns:
        float: returns input value if check passed

    Raises:
        ValueError: Value should be between {minimum} and {maximum}, got {value}.
    """
    if value < minimum or value > maximum:
        raise ValueError(f"Value should be between {minimum} and {maximum}, got {value}.")
    return value


def is_new_api(response: Union[dict, GrowList]) -> bool:
    """
    Checks whether the response came from the new API endpoint or the old API endpoint. The new endpoint response has
    2 differences: it includes a "crs" item in the response geojson and each feature contains a new property called
    'GmlID'. This function checks for these differences and returns a boolean.
    Args:
        response (Union[dict, GrowList]): response from the API. This could take any of 3 forms: the raw json response,
        a single feature, or a GrowList containing features.

    Returns (bool): True if response came from new endpoint, False otherwise
    """
    if isinstance(response, GrowList) and response:
        response = response.values[0]
    if "features" in response:
        return True if "crs" in response else False
    elif "geometry" in response and "properties" in response:
        return True if "GmlID" in response["properties"] else False
    else:
        raise ValueError("Unknown input. Must be either a FeatureCollection, a Feature as a dict, or a GrowList"
                         "containing features as dicts")



from collections.abc import Iterable
from typing import Union

import requests
from typeguard import typechecked

import osdatahub
from osdatahub import Extent
from osdatahub.grow_list import GrowList
from osdatahub.utils import addresses_to_geojson, validate_in_range
from osdatahub.codes import DATASET

class PlacesAPI:
    """Main class for querying the OS Places API (https://osdatahub.os.uk/docs/places/overview)

    Args:
        key (str): A valid OS API Key. Get a free key here - https://osdatahub.os.uk/

    Example::

        from osdatahub import PlacesAPI, Extent
        from os import environ

        key = environ.get("OS_API_KEY")
        places = PlacesAPI(key)
        extent = Extent.from_radius((437293, 115515), 800, "EPSG:27700")
        results = places.query(extent, limit=100)
    """

    __ENDPOINT = r"https://api.os.uk/search/places/v1/"
    HEADERS = {"method": "POST", "headers": "{'Content-Type': 'application/json'}"}

    def __init__(self, key: str):
        self.key = key

    def __endpoint(self, api_name: str) -> str:
        return self.__ENDPOINT + api_name + f"?key={self.key}"

    @staticmethod
    def __get_dataset_param(dataset: Union[str, Iterable] ) -> str:
        if not isinstance(dataset, str):
            dataset_unique = set(dataset)
            shared_datasets = dataset_unique & DATASET

            if len(shared_datasets) == len(dataset_unique):
                return ",".join(dataset_unique)

        elif  dataset in DATASET:
            return dataset

        raise ValueError(f"Unrecognised dataset, expected 'LPI', 'DPA' or ['LPI', 'DPA'], got {dataset}")

    @typechecked
    def query(
            self,
            extent: Extent,
            output_crs: Union[str, None] = None,
            limit: int = 100,
            classification_code: Union[str, Iterable, None] = None,
            logical_status_code: Union[str, int, None] = None,
            dataset: Union[str, Iterable, None] = None
    ) -> dict:
        """Run a query of the OS Places API within a given extent

        Args:
            extent (Extent): The geographical extent of your query
            output_crs (str, optional): The intended output CRS
            limit (int, optional): The maximum number of features to return.
                Defaults to 100.
            classification_code (str|Iterable[str], optional): Classification codes to filter query by
            logical_status_code (str|int, optional): logical status codes to filter query by
            dataset (str|Iterable, optional): The dataset to return. Multiple values can be sent, separated by a comma. Default: DPA.

        Returns:
            FeatureCollection: The results of the query in GeoJSON format
        """
        if not output_crs:
            output_crs = extent.crs
        data = GrowList()
        params = {
            "url": self.__endpoint("polygon"),
            "headers": self.HEADERS,
            "json": extent.to_json(),
            "params": {"srs": extent.crs, "output_srs": output_crs},
        }
        if classification_code or logical_status_code:
            params["params"].update(
                {"fq": self.__format_fq(classification_code, logical_status_code)}
            )

        if dataset is not None:
            params["params"].update(
                {"dataset": self.__get_dataset_param(dataset)}
            )

        try:
            n_required = min(limit, 100)
            while n_required > 0 and data.grown:
                params["params"].update({"offset": len(data), "maxresults": n_required})
                response = osdatahub.post(**params)
                data.extend(self.__format_response(response))
                n_required = min(100, limit - len(data))
        except KeyError:
            response.raise_for_status()
        return addresses_to_geojson(data.values, output_crs)

    @typechecked
    def find(
            self,
            text: str,
            output_crs: str = "EPSG:27700",
            limit: int = 100,
            classification_code: Union[str, Iterable, None] = None,
            logical_status_code: Union[str, int, None] = None,
            minmatch: Union[float, None] = None,
            matchprecision: Union[int, None] = None,
            dataset: Union[str, Iterable, None] = None
    ) -> dict:
        """A free text query of the OS Places API

        Args:
            text (str): The free text search parameter
            output_crs (str, optional): The intended output CRS
            limit (int, optional): The maximum number of features to return.
                Defaults to 100.
            classification_code (str|Iterable[str], optional): Classification codes to filter query by
            logical_status_code (str|int, optional): logical status codes to filter query by
            minmatch (float, optional): The minimum match score a result has to have to be returned
            matchprecision (int, optional): The decimal point position at which the match score value is to be truncated
            dataset (str|Iterable, optional): The dataset to return. Multiple values can be sent, separated by a comma. Default: DPA.

        Returns:
            FeatureCollection: The results of the query in GeoJSON format
        """
        data = GrowList()
        params = {"query": text, "output_srs": output_crs}
        if minmatch is not None:
            params["minmatch"] = validate_in_range(minmatch, 0.1, 1)
        if matchprecision is not None:
            params["matchprecision"] = str(validate_in_range(matchprecision, 1, 10))
        if classification_code or logical_status_code:
            params.update(
                {"fq": self.__format_fq(classification_code, logical_status_code)}
            )
        
        if dataset is not None:
            params.update(
                {"dataset": self.__get_dataset_param(dataset)}
            )

        try:
            n_required = min(limit, 100)
            while n_required > 0 and data.grown:
                params.update({"offset": len(data), "maxresults": n_required})
                response = osdatahub.get(self.__endpoint("find"), params=params, proxies=osdatahub.get_proxies())
                data.extend(self.__format_response(response))
                n_required = min(100, limit - len(data))
        except KeyError:
            response.raise_for_status()
        return addresses_to_geojson(data.values, output_crs)

    @typechecked
    def postcode(
            self,
            postcode: str,
            output_crs: str = "EPSG:27700",
            limit: int = 100,
            classification_code: Union[str, Iterable, None] = None,
            logical_status_code: Union[str, int, None] = None,
            dataset: Union[str, Iterable, None] = None
    ) -> dict:
        """A query based on a property’s postcode. The minimum for the
        resource is the area and district

        e.g. SO16, and will accept a full postcode consisting of the area,
        district, sector and unit e.g. SO16 0AS

        Args:
            postcode (str): The postcode search parameter
            output_crs (str, optional): The intended output CRS.
                Defaults to "EPSG:27700".
            limit (int, optional): The maximum number of features to return.
                Defaults to 100.
            classification_code (str|Iterable[str], optional): Classification codes to filter query by
            logical_status_code (str|int, optional): logical status codes to filter query by
            dataset (str|Iterable, optional): The dataset to return. Multiple values can be sent, separated by a comma. Default: DPA.

        Returns:
            FeatureCollection: The results of the query in GeoJSON format
        """
        data = GrowList()
        params = {"postcode": postcode, "output_srs": output_crs}
        if classification_code or logical_status_code:
            params.update(
                {"fq": self.__format_fq(classification_code, logical_status_code)}
            )
        if dataset is not None:
            params.update(
                {"dataset": self.__get_dataset_param(dataset)}
            )

        try:
            n_required = min(limit, 100)
            while n_required > 0 and data.grown:
                params.update({"offset": len(data), "maxresults": n_required})
                response = osdatahub.get(self.__endpoint("postcode"), params=params, proxies=osdatahub.get_proxies())
                data.extend(self.__format_response(response))
                n_required = min(100, limit - len(data))
        except KeyError:
            response.raise_for_status()
        return addresses_to_geojson(data.values, output_crs)

    @typechecked
    def uprn(
            self,
            uprn: int,
            output_crs: str = "EPSG:27700",
            classification_code: Union[str, Iterable, None] = None,
            logical_status_code: Union[str, int, None] = None,
            dataset: Union[str, Iterable, None] = None
    ) -> dict:
        """A query that takes a UPRN as the search parameter

        Args:
            uprn (int): A Valid UPRN
            output_crs (str, optional): The intended output CRS.
                Defaults to "EPSG:27700".
            classification_code (str|Iterable[str], optional): Classification codes to filter query by
            logical_status_code (str|int, optional): logical status codes to filter query by
            dataset (str|Iterable, optional): The dataset to return. Multiple values can be sent, separated by a comma. Default: DPA.

        Returns:
            FeatureCollection: The results of the query in GeoJSON format
        """
        data = GrowList()
        params = {"uprn": uprn, "output_srs": output_crs}
        if classification_code or logical_status_code:
            params.update(
                {"fq": self.__format_fq(classification_code, logical_status_code)}
            )
        if dataset is not None:
            params.update(
                {"dataset": self.__get_dataset_param(dataset)}
            )
        try:
            response = osdatahub.get(self.__endpoint("uprn"), params=params, proxies=osdatahub.get_proxies())
            data.extend(self.__format_response(response))
        except KeyError:
            response.raise_for_status()
        return addresses_to_geojson(data.values, output_crs)

    @typechecked
    def nearest(
            self,
            point: tuple,
            point_crs: str,
            radius: float = 100,
            output_crs: str = "EPSG:27700",
            classification_code: Union[str, Iterable] = None,
            logical_status_code: Union[str, int] = None,
            dataset: Union[str, Iterable, None] = None
    ) -> dict:
        """Takes a pair of coordinates (X, Y)/(Lon, Lat) as an input
        to determine the closest address.

        Args:
            point (tuple): A set of coordinates
            point_crs (str): The crs corresponding to the point coordinates
            radius (float): The search radius in metres (max. 1000).
                Defaults to 100.
            output_crs (str, optional): The intended output CRS.
                Defaults to "EPSG:27700".
            classification_code (str|Iterable[str], optional): Classification codes to filter query by
            logical_status_code (str|int, optional): logical status codes to filter query by
            dataset (str|Iterable, optional): The dataset to return. Multiple values can be sent, separated by a comma. Default: DPA.

        Returns:
            FeatureCollection: The results of the query in GeoJSON format
        """
        data = GrowList()
        point = point if point_crs.upper() != "EPSG:4326" else (point[1], point[0])
        params = {
            "point": ",".join([str(c) for c in point]),
            "srs": point_crs,
            "output_srs": output_crs,
            "radius": radius,
        }
        if classification_code or logical_status_code:
            params.update(
                {"fq": self.__format_fq(classification_code, logical_status_code)}
            )
        if dataset is not None:
            params.update(
                {"dataset": self.__get_dataset_param(dataset)}
            )
        try:
            response = osdatahub.get(self.__endpoint("nearest"), params=params, proxies=osdatahub.get_proxies())
            data.extend(self.__format_response(response))
        except KeyError:
            response.raise_for_status()
        return addresses_to_geojson(data.values, output_crs)

    @staticmethod
    def __format_response(response: requests.Response) -> list:
        results = response.json()["results"]
        return [result[list(result.keys())[0]] for result in results]

    @staticmethod
    @typechecked
    def __format_fq(
            classification_code: Union[str, Iterable, None] = None,
            logical_status_code: Union[str, int, None] = None,
    ) -> list:
        """
        Formats optional fq arguments for Places API query

        Args:
            classification_code (str|Iterable[str], optional): The classification codes to filter query
            logical_status_code (str|Number, optional): Logical status code to filter query

        Returns:
            list of fq filtering arguments
        """
        fq_args = []
        if classification_code:
            if isinstance(classification_code, str):
                class_codes = "classification_code:" + classification_code
            elif isinstance(classification_code, Iterable):
                class_codes = " ".join(
                    [f"classification_code:{arg}" for arg in classification_code]
                )
            else:
                raise TypeError(
                    f"'classification_code' argument must be Iterable or str, but was type {type(classification_code)}"
                )

            fq_args.append(class_codes)
        if logical_status_code:
            if not str(logical_status_code).isnumeric():
                raise TypeError(
                    "logical_status_code can have a maximum of 1 filter and must have a numeric value."
                )

            fq_args.append("logical_status_code:" + str(logical_status_code))

        return fq_args


if __name__ == "__main__":
    from os import environ

    key = environ.get("OS_API_KEY")
    places = PlacesAPI(key)

    # extent = Extent.from_bbox((-3.2939550711619177,50.746391786819316,-3.2788419310229244,50.75566426785872), "EPSG:4326")
    # results = places.query(extent, limit=42)
    results = places.find(
        "Ordnance Survey Adanac Drive SO16",
        minmatch=0.4,
        matchprecision=9
    )
    # places.nearest((-3.2939550711619177,50.746391786819316), "EPSG:4326", 1000)

    print(results)
    # import json

    # with open("test.json", "w") as f:
    #     json.dump(results, f)



import pytest
from osdatahub import LinkedIdentifiersAPI
from osdatahub.LinkedIdentifiersAPI.linked_identifier_options import Options

from tests.data import linked_identifiers_api_data as data


class TestLinkedIdentifiersAPI:
    @pytest.mark.parametrize(*data.test_get_endpoint())
    def test_get_endpoint(
            self, key, identifier, feature_type, identifier_type, expected_result
    ):
        # Arrange
        li_api = LinkedIdentifiersAPI(key)

        endpoint = li_api._LinkedIdentifiersAPI__get_endpoint(
            identifier, feature_type, identifier_type
        )

        # Assert
        assert endpoint == expected_result

    def test_get_endpoint_error(self):
        # Arrange
        li_api = LinkedIdentifiersAPI("KEY")

        with pytest.raises(ValueError) as exc_info:
            li_api._LinkedIdentifiersAPI__get_endpoint(
                "ID", "featureType", "identifierType"
            )
        # Assert
        assert (
                str(exc_info.value)
                == "It is possible to query by the feature_type "
                + "OR the identifier type, but not both"
        )


class TestOptions:
    def test_validation_error(self):
        # Arrange
        options = Options("test options", "option1 option2 option3".split())

        with pytest.raises(ValueError) as exc_info:
            options.validate("option4")

        # Assert
        assert str(exc_info.value) == (
                f"'option4' is not a valid test options, "
                + f"please choose from one of the following:\n- option1\n- option2\n- option3"
        )



"""
Script that queries all places and features endpoints that are implemented 
in osdatahub.

This is just for convenience and is intentionally left unrecognisable to 
pytest. Otherwise, running tests could cost developers real money!
"""

from os import environ

from osdatahub import Extent, FeaturesAPI, PlacesAPI, LinkedIdentifiersAPI

# Get API key
key = environ.get("OS_API_KEY")

# Set extent
extent = Extent.from_bbox((600000, 310200, 600900, 310900), "EPSG:27700")

print(f"{'Running':=^40}")

"""
======= FeaturesAPI =======
"""

features = FeaturesAPI(key, "zoomstack_local_buildings", extent)
results = features.query(limit=1)

assert (
        len(results["features"]) == 1
), f"FeaturesAPI.query() FAILED: {len(results['features']) = }"
print(f"FeaturesAPI.query(){'PASSED':.>20}")

"""
======= PlacesAPI =======
"""

# query
places = PlacesAPI(key)
results = places.query(extent, limit=1)

assert (
        len(results["features"]) == 1
), f"PlacesAPI.query() FAILED: {len(results['features']) = }"
print(f"PlacesAPI.query(){'PASSED':.>20}")

# find
places = PlacesAPI(key)
results = places.find(
    "test", classification_code=("CI04"), logical_status_code=1, limit=1
)

assert (
        len(results["features"]) == 1
), f"PlacesAPI.find() failed: {len(results['features']) = }"
print(f"PlacesAPI.find(){'PASSED':.>20}")

# postcode
places = PlacesAPI(key)
results = places.postcode(
    "SO21", classification_code=("RD02", "RD04"), logical_status_code=1, limit=1
)

assert (
        len(results["features"]) == 1
), f"PlacesAPI.postcode() FAILED: {len(results['features']) = }"
print(f"PlacesAPI.postcode(){'PASSED':.>20}")

# uprn
places = PlacesAPI(key)
results = places.uprn(100062015780)

assert (
        len(results["features"]) == 1
), f"PlacesAPI.uprn() FAILED: {len(results['features']) = }"
print(f"PlacesAPI.uprn(){'PASSED':.>20}")

# nearest
places = PlacesAPI(key)
results = places.nearest((600000, 310200), "EPSG:27700", radius=500)

assert (
        len(results["features"]) == 1
), f"PlacesAPI.nearest() FAILED: {len(results['features']) = }"
print(f"PlacesAPI.nearest(){'PASSED':.>20}")

"""
======= LinkedIdentifiersAPI =======
"""

# query id
linked_id = LinkedIdentifiersAPI(key)
results = linked_id.query(80833669)

assert (
        len(results["linkedIdentifiers"]) == 1
), f"LinkedIdentifiersAPI.query() FAILED: {len(results['linkedIdentifiers']) = }"
print(f"LinkedIdentifiersAPI.query(){'PASSED':.>20}")

# query id and feature_type
linked_id = LinkedIdentifiersAPI(key)
results = linked_id.query(80833669, feature_type="Street")

assert (
        len(results["linkedIdentifier"]) == 3
), f"LinkedIdentifiersAPI.query(feature_type) FAILED: {len(results['linkedIdentifier']) = }"
print(f"LinkedIdentifiersAPI.query(feature_type){'PASSED':.>20}")

# query id and identifier_type
linked_id = LinkedIdentifiersAPI(key)
results = linked_id.query(80833669, identifier_type="USRN")

assert (
        len(results["linkedIdentifier"]) == 3
), f"LinkedIdentifiersAPI.query(identifier_type) FAILED: {len(results['linkedIdentifier']) = }"
print(f"LinkedIdentifiersAPI.query(identifier_type){'PASSED':.>20}")

print(f"{'Complete':=^40}")



"""Tests for the async NGD API client."""

import os
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

import osdatahub
from osdatahub import Extent
from osdatahub.AsyncAPI import AsyncHTTPClient
from osdatahub.NGD.async_ngd_api import AsyncNGD

API_KEY = os.environ.get("OSDATAHUB_TEST_KEY")


@pytest.mark.asyncio
async def test_async_ngd_query():
    """Test basic async NGD query functionality."""
    response = {
        "type": "FeatureCollection",
        "features": [{"id": 1}, {"id": 2}],
        "numberReturned": 2,
        "links": [],
    }

    with patch.object(AsyncHTTPClient, "get", new_callable=AsyncMock) as mock_get:
        mock_get.return_value = response

        async with AsyncNGD("test-key", "test-collection", request_delay=0) as ngd:
            result = await ngd.query(max_results=10)

        assert result.numberReturned == 2
        assert len(result.features) == 2


@pytest.mark.asyncio
@pytest.mark.skipif(not API_KEY, reason="Test API key not available")
@pytest.mark.parametrize(
    "collection",
    [
        "trn-ntwk-roadlink-4",
        "trn-ntwk-street-1",
    ],
)
async def test_async_ngd_query_real_data(collection):
    """Test async NGD query with real API."""
    assert API_KEY is not None
    extent = Extent.from_bbox((436000, 114000, 438000, 116000), "EPSG:27700")

    async with AsyncNGD(
        key=API_KEY,
        collection=collection,
        max_concurrent=5,
        request_delay=0.02,
    ) as ngd:
        result = await ngd.query(extent=extent, max_results=None)

    assert result.numberReturned > 0
    assert len(result.features) == result.numberReturned


class TestAsyncHTTPClientProxy:
    """Tests for AsyncHTTPClient proxy support."""

    def test_get_proxy_https(self):
        """Test that HTTPS URLs use the https proxy."""
        client = AsyncHTTPClient(proxies={"https": "http://proxy:8080"})
        assert client._get_proxy("https://example.com") == "http://proxy:8080"

    def test_get_proxy_http(self):
        """Test that HTTP URLs use the http proxy."""
        client = AsyncHTTPClient(proxies={"http": "http://proxy:8080"})
        assert client._get_proxy("http://example.com") == "http://proxy:8080"

    def test_get_proxy_empty(self):
        """Test that empty proxies dict returns None."""
        client = AsyncHTTPClient(proxies={})
        assert client._get_proxy("https://example.com") is None

    def test_get_proxy_none(self):
        """Test that no proxies parameter returns None."""
        client = AsyncHTTPClient()
        assert client._get_proxy("https://example.com") is None

    def test_get_proxy_both(self):
        """Test proxy selection with both http and https configured."""
        client = AsyncHTTPClient(
            proxies={
                "http": "http://http-proxy:8080",
                "https": "http://https-proxy:8080",
            }
        )
        assert client._get_proxy("https://example.com") == "http://https-proxy:8080"
        assert client._get_proxy("http://example.com") == "http://http-proxy:8080"


class TestAsyncNGDProxy:
    """Tests for AsyncNGD proxy configuration."""

    def test_async_ngd_uses_proxies(self):
        """Test that AsyncNGD passes proxies to the HTTP client."""
        test_proxies = {"https": "http://proxy:8080"}

        with patch.object(osdatahub, "get_proxies", return_value=test_proxies):
            ngd = AsyncNGD("test-key", "test-collection")
            client = ngd._get_client()
            assert client._proxies == test_proxies

    def test_async_ngd_empty_proxies(self):
        """Test that AsyncNGD works with empty proxies (default)."""
        with patch.object(osdatahub, "get_proxies", return_value={}):
            ngd = AsyncNGD("test-key", "test-collection")
            client = ngd._get_client()
            assert client._proxies == {}

    @pytest.mark.asyncio
    async def test_async_ngd_get_collections_uses_proxies(self):
        """Test that get_collections passes proxies to the HTTP client."""
        test_proxies = {"https": "http://proxy:8080"}

        AsyncNGD.clear_collections_cache()

        with patch.object(osdatahub, "get_proxies", return_value=test_proxies):
            with patch.object(
                AsyncHTTPClient, "get", new_callable=AsyncMock
            ) as mock_get:
                with patch.object(AsyncHTTPClient, "close", new_callable=AsyncMock):
                    mock_get.return_value = {"collections": []}
                    await AsyncNGD.get_collections()

        AsyncNGD.clear_collections_cache()

    @pytest.mark.asyncio
    async def test_proxy_passed_to_aiohttp(self):
        """Test that proxy is actually passed to aiohttp session.get()."""
        test_proxies = {"https": "http://proxy:8080"}

        with patch.object(osdatahub, "get_proxies", return_value=test_proxies):
            with patch("aiohttp.ClientSession.get") as mock_session_get:
                mock_response = AsyncMock()
                mock_response.json = AsyncMock(
                    return_value={"features": [], "numberReturned": 0, "links": []}
                )
                mock_response.raise_for_status = MagicMock()
                mock_response.headers = {}
                mock_session_get.return_value.__aenter__.return_value = mock_response

                async with AsyncNGD("key", "collection", request_delay=0) as ngd:
                    await ngd.query(max_results=10)

                mock_session_get.assert_called()
                call_kwargs = mock_session_get.call_args.kwargs
                assert call_kwargs.get("proxy") == "http://proxy:8080"



import json

from osdatahub import Extent
from osdatahub.bbox import BBox
from pytest import param
from shapely.geometry import box, Polygon, Point


def test_from_bbox():
    test_variables = "bbox, crs, expected_result"
    test_data = [
        param(
            (600000, 310200, 600900, 310800),
            "epsg:27700",
            Extent(box(600000, 310200, 600900, 310800), "epsg:27700"),
            id="tuple bbox",
        ),
        param(
            BBox(600000, 310200, 600900, 310800),
            "epsg:27700",
            Extent(box(600000, 310200, 600900, 310800), "epsg:27700"),
            id="bbox object",
        ),
    ]
    return test_variables, test_data


def test_from_radius():
    test_variables = "centre, radius, crs, expected_result, expected_bbox"
    test_data = [
        param(
            (0, 0),
            10,
            "epsg:27700",
            Extent(Point(0, 0).buffer(10), "epsg:27700"),
            BBox(-10, -10, 10, 10),
            id="tuple centre",
        ),
        param(
            Point(0, 0),
            10,
            "epsg:27700",
            Extent(Point(0, 0).buffer(10), "epsg:27700"),
            BBox(-10, -10, 10, 10),
            id="shapely point centre",
        ),
        param(
            [0, 0],
            10,
            "epsg:27700",
            Extent(Point(0, 0).buffer(10), "epsg:27700"),
            BBox(-10, -10, 10, 10),
            id="list centre",
        ),
    ]
    return test_variables, test_data


def test_set_crs():
    test_variables = "initital_polygon, initital_crs, updated_crs, expected_result"
    test_data = [
        param(
            Polygon(),
            "epsg:27700",
            "epsg:3857",
            Extent(Polygon(), "epsg:3857"),
            id="set crs",
        )
    ]
    return test_variables, test_data


def test_is_within():
    test_variables = "polygon, crs, bounds, expected_result"
    test_data = [
        param(
            Polygon([(0, 0), (5, 5), (10, 0)]),
            "epsg:27700",
            (0, 0, 10, 5),
            True,
            id="tuple bbox - is equal to bounds",
        ),
        param(
            Polygon([(0, 0), (5, 5), (10, 0)]),
            "epsg:27700",
            (0, 0, 1, 5),
            False,
            id="tuple bbox - is not within",
        ),
        param(
            Polygon([(0, 0), (5, 5), (10, 0)]),
            "epsg:27700",
            (-1, -1, 15, 15),
            True,
            id="tuple bbox - is fully within",
        ),
        param(
            Polygon([(0, 0), (5, 5), (10, 0)]),
            "epsg:27700",
            BBox(-1, -1, 15, 15),
            True,
            id="BBox object - is fully within",
        ),
    ]
    return test_variables, test_data


def test_xml_coords():
    test_variables = "polygon, crs, expected_result"
    test_data = [
        param(
            Polygon([(0, 0), (5, 5), (10, 0)]),
            "epsg:27700",
            "0.0,0.0 5.0,5.0 10.0,0.0 0.0,0.0",
            id="triangle",
        ),
        param(
            Polygon([(0, 0), (5, 7), (10, 0)]),
            "epsg:27700",
            "0.0,0.0 5.0,7.0 10.0,0.0 0.0,0.0",
            id="triangle",
        ),
        param(
            Polygon([(-3.29, 50.74), (-3.27, 50.74), (-3.27, 50.76)]),
            "epsg:4326",
            "50.74,-3.29 50.74,-3.27 50.76,-3.27 50.74,-3.29",
            id="bbox object - lonlat input",
        ),
    ]
    return test_variables, test_data


def test_from_ons_code():
    test_variables = "ons_code, ons_mock_response, expected_result"
    test_data = [
        param(
            "E0100001",
            json.dumps(
                {
                    "type": "FeatureCollection",
                    "features": [
                        {
                            "type": "Feature",
                            "geometry": {
                                "type": "Polygon",
                                "coordinates": [[[0, 0], [1, 1], [2, 0]]],
                            },
                        }
                    ],
                }
            ),
            Extent(Polygon([(0, 0), (1, 1), (2, 0)]), "EPSG:4326"),
            id="triangle",
        )
    ]
    return test_variables, test_data


def test_from_ons_code_error():
    test_variables = "ons_code, ons_mock_response"
    test_data = [
        param(
            "E0100001",
            json.dumps(
                {
                    "type": "FeatureCollection",
                    "features": [
                        {
                            "type": "Feature",
                            "geometry": {
                                "type": "LineString",
                                "coordinates": [[[0, 0], [1, 1], [2, 0]]],
                            },
                        }
                    ],
                }
            ),
            id="line",
        )
    ]
    return test_variables, test_data


def test_to_json():
    test_variables = "polygon, crs, expected_result"
    test_data = [
        param(
            Polygon([(0, 0), (5, 5), (10, 0)]),
            "epsg:27700",
            {"type": "Polygon", "coordinates": [[(0, 0), (5, 5), (10, 0), (0, 0)]]},
            id="triangle - BNG",
        ),
        param(
            Polygon([(-3.29, 50.74), (-3.27, 50.74), (-3.27, 50.76)]),
            "epsg:4326",
            {
                "type": "Polygon",
                "coordinates": [
                    [(50.74, -3.29), (50.74, -3.27), (50.76, -3.27), (50.74, -3.29)]
                ],
            },
            id="triangle - lonlat",
        ),
    ]
    return test_variables, test_data



from typing import Union, Iterable

LOGICAL_STATUS_CODES = {
    1: "Approved address",
    3: "Alternate address",
    6: "Provisional address",
    8: "Historic address"
}


def validate_logical_status_code(logical_status_code: Union[str, int]) -> bool:
    return logical_status_code in LOGICAL_STATUS_CODES.keys()


COUNTRY_CODES = {
    "E": "This record is within England",
    "W": "This record is within Wales",
    "S": "This record is within Scotland",
    "N": "This record is within Northern Island",
    "L": "This record is within Channel Islands",
    "M": "This record is within the Isle of Man",
    "J": "This record is not assigned to a country as it falls outside of the land boundaries used"
}

AREA_CODES = {"GB", "HP", "HT", "HU", "HW", "HX", "HY", "HZ", "NA", "NB", "NC", "ND", "NF", "NG", "NH", "NJ", "NK",
              "NL", "NM", "NN," "NO", "NR", "NS", "NT", "NU", "NW", "NX", "NY", "NZ", "OV", "SD", "SE", "TA", "SH",
              "SJ", "SK", "TF", "TG", "SM", "SN", "SO", "SP", "TL", "TM", "SR", "SS", "ST", "SU", "TQ", "TR", "SV",
              "SW", "SX", "SY", "SZ", "TV"}


def validate_country_codes(country_code: Union[Iterable, str]) -> set:
    country_code = set(country_code) if isinstance(country_code, Iterable) else {country_code}
    return country_code - set(COUNTRY_CODES.keys())


DATASET = {
    "DPA",
    "LPI"
}



from pytest import param

geojson_1_feature = {
    "type": "FeatureCollection",
    "links": [
        {
            "href": "https://api.os.uk/features/ngd/ofa/v1/collections/bld-fts-buildingline/items?limit=1",
            "rel": "self",
            "type": "application/geo+json",
            "title": "All features from the 'Building Line' collection"
        },
        {
            "href": "https://api.os.uk/features/ngd/ofa/v1/collections/bld-fts-buildingline/items?offset=1&limit=1",
            "rel": "next",
            "type": "application/geo+json",
            "title": "Next page"
        }
    ],
    "timeStamp": "2022-10-28T14:20:26.387528Z",
    "numberReturned": 1,
    "features": [
        {
            "id": "00000016-e0a2-45ca-855a-645753d72716",
            "type": "Feature",
            "geometry": {
                "type": "LineString",
                "coordinates": [
                    [
                        -3.8649573,
                        51.8129786
                    ],
                    [
                        -3.8649033,
                        51.8129757
                    ]
                ]
            },
            "properties": {
                "osid": "00000016-e0a2-45ca-855a-645753d72716",
                "toid": "osgb5000005271044468",
                "theme": "Buildings",
                "changetype": "New",
                "isobscured": "false",
                "description": "Building Internal Division",
                "versiondate": "2022-08-26",
                "physicallevel": "Surface Level",
                "geometry_length": 3.734355,
                "geometry_source": "Ordnance Survey",
                "description_source": "Ordnance Survey",
                "geometry_updatedate": "2020-09-23",
                "capturespecification": "Rural",
                "geometry_evidencedate": "2019-07-05",
                "description_updatedate": "2020-09-23",
                "versionavailabletodate": "null",
                "firstdigitalcapturedate": "2020-09-23",
                "description_evidencedate": "2019-07-05",
                "versionavailablefromdate": "2022-08-27T00:00:00Z"
            }
        }
    ]
}

geojson_10_features = {
    "type": "FeatureCollection",
    "links": [
        {
            "href": "https://api.os.uk/features/ngd/ofa/v1/collections/bld-fts-buildingline/items?limit=10",
            "rel": "self",
            "type": "application/geo+json",
            "title": "All features from the 'Building Line' collection"
        },
        {
            "href": "https://api.os.uk/features/ngd/ofa/v1/collections/bld-fts-buildingline/items?offset=10&limit=10",
            "rel": "next",
            "type": "application/geo+json",
            "title": "Next page"
        }
    ],
    "timeStamp": "2022-10-28T14:21:21.524232Z",
    "numberReturned": 10,
    "features": [
        {
            "id": "00000016-e0a2-45ca-855a-645753d72716",
            "type": "Feature",
            "geometry": {
                "type": "LineString",
                "coordinates": [
                    [
                        -3.8649573,
                        51.8129786
                    ],
                    [
                        -3.8649033,
                        51.8129757
                    ]
                ]
            },
            "properties": {
                "osid": "00000016-e0a2-45ca-855a-645753d72716",
                "toid": "osgb5000005271044468",
                "theme": "Buildings",
                "changetype": "New",
                "isobscured": "false",
                "description": "Building Internal Division",
                "versiondate": "2022-08-26",
                "physicallevel": "Surface Level",
                "geometry_length": 3.734355,
                "geometry_source": "Ordnance Survey",
                "description_source": "Ordnance Survey",
                "geometry_updatedate": "2020-09-23",
                "capturespecification": "Rural",
                "geometry_evidencedate": "2019-07-05",
                "description_updatedate": "2020-09-23",
                "versionavailabletodate": "null",
                "firstdigitalcapturedate": "2020-09-23",
                "description_evidencedate": "2019-07-05",
                "versionavailablefromdate": "2022-08-27T00:00:00Z"
            }
        },
        {
            "id": "00000052-e07e-4420-b2d5-b6ff90afcac2",
            "type": "Feature",
            "geometry": {
                "type": "LineString",
                "coordinates": [
                    [
                        -3.0641844,
                        50.9584926
                    ],
                    [
                        -3.0643669,
                        50.9585053
                    ]
                ]
            },
            "properties": {
                "osid": "00000052-e07e-4420-b2d5-b6ff90afcac2",
                "toid": "osgb1000000283880279",
                "theme": "Buildings",
                "changetype": "New",
                "isobscured": "false",
                "description": "Overhanging Building Edge",
                "versiondate": "2022-08-26",
                "physicallevel": "Level 1",
                "geometry_length": 12.899612,
                "geometry_source": "Ordnance Survey",
                "description_source": "Ordnance Survey",
                "geometry_updatedate": "2004-01-28",
                "capturespecification": "Rural",
                "geometry_evidencedate": "2004-01-28",
                "description_updatedate": "2004-01-28",
                "versionavailabletodate": "null",
                "firstdigitalcapturedate": "1992-02-01",
                "description_evidencedate": "2004-01-28",
                "versionavailablefromdate": "2022-08-27T00:00:00Z"
            }
        },
        {
            "id": "000000b7-a66f-42b5-870b-3ea9729b73e3",
            "type": "Feature",
            "geometry": {
                "type": "LineString",
                "coordinates": [
                    [
                        -1.2987328,
                        54.0082184
                    ],
                    [
                        -1.2987991,
                        54.0081463
                    ]
                ]
            },
            "properties": {
                "osid": "000000b7-a66f-42b5-870b-3ea9729b73e3",
                "toid": "osgb1000000080069326",
                "theme": "Buildings",
                "changetype": "New",
                "isobscured": "false",
                "description": "Building Internal Division",
                "versiondate": "2022-08-26",
                "physicallevel": "Surface Level",
                "geometry_length": 9.130049,
                "geometry_source": "Ordnance Survey",
                "description_source": "Ordnance Survey",
                "geometry_updatedate": "2007-02-20",
                "capturespecification": "Rural",
                "geometry_evidencedate": "2007-02-20",
                "description_updatedate": "2007-02-20",
                "versionavailabletodate": "null",
                "firstdigitalcapturedate": "1992-09-24",
                "description_evidencedate": "2007-02-20",
                "versionavailablefromdate": "2022-08-27T00:00:00Z"
            }
        },
        {
            "id": "00000102-d65c-4396-8635-6f0a049023d9",
            "type": "Feature",
            "geometry": {
                "type": "LineString",
                "coordinates": [
                    [
                        -1.9501576,
                        52.5456803
                    ],
                    [
                        -1.9501459,
                        52.5456862
                    ],
                    [
                        -1.95005,
                        52.5457325
                    ]
                ]
            },
            "properties": {
                "osid": "00000102-d65c-4396-8635-6f0a049023d9",
                "toid": "osgb1000000124252224",
                "theme": "Buildings",
                "changetype": "New",
                "isobscured": "false",
                "description": "Building Occupier Division",
                "versiondate": "2022-08-26",
                "physicallevel": "Surface Level",
                "geometry_length": 9.324264,
                "geometry_source": "Ordnance Survey",
                "description_source": "Ordnance Survey",
                "geometry_updatedate": "2016-08-01",
                "capturespecification": "Urban",
                "geometry_evidencedate": "2016-04-20",
                "description_updatedate": "1983-10-13",
                "versionavailabletodate": "null",
                "firstdigitalcapturedate": "1983-10-13",
                "description_evidencedate": "1983-10-13",
                "versionavailablefromdate": "2022-08-27T00:00:00Z"
            }
        },
        {
            "id": "0000013e-5fed-447d-a627-dae6fb215138",
            "type": "Feature",
            "geometry": {
                "type": "LineString",
                "coordinates": [
                    [
                        -3.3809754,
                        52.2413303
                    ],
                    [
                        -3.3809517,
                        52.2413255
                    ]
                ]
            },
            "properties": {
                "osid": "0000013e-5fed-447d-a627-dae6fb215138",
                "toid": "osgb1000000264266704",
                "theme": "Buildings",
                "changetype": "New",
                "isobscured": "false",
                "description": "Building Internal Division",
                "versiondate": "2022-08-26",
                "physicallevel": "Surface Level",
                "geometry_length": 1.707923,
                "geometry_source": "Ordnance Survey",
                "description_source": "Ordnance Survey",
                "geometry_updatedate": "2005-06-22",
                "capturespecification": "Rural",
                "geometry_evidencedate": "2005-06-22",
                "description_updatedate": "2005-06-22",
                "versionavailabletodate": "null",
                "firstdigitalcapturedate": "1992-04-13",
                "description_evidencedate": "2005-06-22",
                "versionavailablefromdate": "2022-08-27T00:00:00Z"
            }
        },
        {
            "id": "0000031f-f2d0-4547-8d91-715e8073a00b",
            "type": "Feature",
            "geometry": {
                "type": "LineString",
                "coordinates": [
                    [
                        -0.4105347,
                        53.7419995
                    ],
                    [
                        -0.4105707,
                        53.7420023
                    ]
                ]
            },
            "properties": {
                "osid": "0000031f-f2d0-4547-8d91-715e8073a00b",
                "toid": "osgb1000000012849650",
                "theme": "Buildings",
                "changetype": "New",
                "isobscured": "false",
                "description": "Building Internal Division",
                "versiondate": "2022-08-26",
                "physicallevel": "Surface Level",
                "geometry_length": 2.39416,
                "geometry_source": "Ordnance Survey",
                "description_source": "Ordnance Survey",
                "geometry_updatedate": "2009-05-29",
                "capturespecification": "Urban",
                "geometry_evidencedate": "2009-05-29",
                "description_updatedate": "2009-05-29",
                "versionavailabletodate": "null",
                "firstdigitalcapturedate": "1999-06-11",
                "description_evidencedate": "2009-05-29",
                "versionavailablefromdate": "2022-08-27T00:00:00Z"
            }
        },
        {
            "id": "0000037a-4ece-460c-add7-5cd68ec700e9",
            "type": "Feature",
            "geometry": {
                "type": "LineString",
                "coordinates": [
                    [
                        0.1697308,
                        51.7160161
                    ],
                    [
                        0.1698108,
                        51.7160542
                    ],
                    [
                        0.1698501,
                        51.7160727
                    ]
                ]
            },
            "properties": {
                "osid": "0000037a-4ece-460c-add7-5cd68ec700e9",
                "toid": "osgb1000000027690091",
                "theme": "Buildings",
                "changetype": "New",
                "isobscured": "false",
                "description": "Building Occupier Division",
                "versiondate": "2022-08-26",
                "physicallevel": "Surface Level",
                "geometry_length": 10.375383,
                "geometry_source": "Ordnance Survey",
                "description_source": "Ordnance Survey",
                "geometry_updatedate": "2021-07-21",
                "capturespecification": "Rural",
                "geometry_evidencedate": "2021-07-21",
                "description_updatedate": "2002-12-11",
                "versionavailabletodate": "null",
                "firstdigitalcapturedate": "1994-06-29",
                "description_evidencedate": "2002-12-11",
                "versionavailablefromdate": "2022-08-27T00:00:00Z"
            }
        },
        {
            "id": "000003f7-455e-4a1f-9691-29c927a3687f",
            "type": "Feature",
            "geometry": {
                "type": "LineString",
                "coordinates": [
                    [
                        -0.1789267,
                        51.3667955
                    ],
                    [
                        -0.1789231,
                        51.3667956
                    ],
                    [
                        -0.1788968,
                        51.3667965
                    ],
                    [
                        -0.1787859,
                        51.366799
                    ]
                ]
            },
            "properties": {
                "osid": "000003f7-455e-4a1f-9691-29c927a3687f",
                "toid": "osgb1000001798457344",
                "theme": "Buildings",
                "changetype": "New",
                "isobscured": "false",
                "description": "Building Occupier Division",
                "versiondate": "2022-08-26",
                "physicallevel": "Surface Level",
                "geometry_length": 9.811248,
                "geometry_source": "Ordnance Survey",
                "description_source": "Ordnance Survey",
                "geometry_updatedate": "2017-09-04",
                "capturespecification": "Urban",
                "geometry_evidencedate": "2016-09-15",
                "description_updatedate": "1970-01-01",
                "versionavailabletodate": "null",
                "firstdigitalcapturedate": "1970-01-01",
                "description_evidencedate": "1970-01-01",
                "versionavailablefromdate": "2022-08-27T00:00:00Z"
            }
        },
        {
            "id": "0000046d-6491-4f76-8d61-a7080b9b83e4",
            "type": "Feature",
            "geometry": {
                "type": "LineString",
                "coordinates": [
                    [
                        -2.198468,
                        52.1985489
                    ],
                    [
                        -2.1984579,
                        52.1985368
                    ]
                ]
            },
            "properties": {
                "osid": "0000046d-6491-4f76-8d61-a7080b9b83e4",
                "toid": "osgb1000002071113695",
                "theme": "Buildings",
                "changetype": "New",
                "isobscured": "false",
                "description": "Overhanging Building Edge",
                "versiondate": "2022-08-26",
                "physicallevel": "Level 1",
                "geometry_length": 1.507216,
                "geometry_source": "Ordnance Survey",
                "description_source": "Ordnance Survey",
                "geometry_updatedate": "1989-10-24",
                "capturespecification": "Urban",
                "geometry_evidencedate": "1989-10-24",
                "description_updatedate": "1989-10-24",
                "versionavailabletodate": "null",
                "firstdigitalcapturedate": "1989-10-24",
                "description_evidencedate": "1989-10-24",
                "versionavailablefromdate": "2022-08-27T00:00:00Z"
            }
        },
        {
            "id": "00000471-3f43-4cfb-a506-9e9c862b113c",
            "type": "Feature",
            "geometry": {
                "type": "LineString",
                "coordinates": [
                    [
                        -1.4225073,
                        53.3766972
                    ],
                    [
                        -1.4224614,
                        53.3767604
                    ]
                ]
            },
            "properties": {
                "osid": "00000471-3f43-4cfb-a506-9e9c862b113c",
                "toid": "osgb1000000107792612",
                "theme": "Buildings",
                "changetype": "New",
                "isobscured": "false",
                "description": "Building Occupier Division",
                "versiondate": "2022-08-26",
                "physicallevel": "Surface Level",
                "geometry_length": 7.661756,
                "geometry_source": "Ordnance Survey",
                "description_source": "Ordnance Survey",
                "geometry_updatedate": "1993-03-01",
                "capturespecification": "Urban",
                "geometry_evidencedate": "1993-03-01",
                "description_updatedate": "1993-03-01",
                "versionavailabletodate": "null",
                "firstdigitalcapturedate": "1993-03-01",
                "description_evidencedate": "1993-03-01",
                "versionavailablefromdate": "2022-08-27T00:00:00Z"
            }
        }
    ]
}

geojson_3_features = {
    "type": "FeatureCollection",
    "links": [
        {
            "href": "https://api.os.uk/features/ngd/ofa/v1/collections/bld-fts-buildingline/items?limit=3",
            "rel": "self",
            "type": "application/geo+json",
            "title": "All features from the 'Building Line' collection"
        },
        {
            "href": "https://api.os.uk/features/ngd/ofa/v1/collections/bld-fts-buildingline/items?offset=3&limit=3",
            "rel": "next",
            "type": "application/geo+json",
            "title": "Next page"
        }
    ],
    "timeStamp": "2022-10-28T14:21:50.334565Z",
    "numberReturned": 3,
    "features": [
        {
            "id": "00000016-e0a2-45ca-855a-645753d72716",
            "type": "Feature",
            "geometry": {
                "type": "LineString",
                "coordinates": [
                    [
                        -3.8649573,
                        51.8129786
                    ],
                    [
                        -3.8649033,
                        51.8129757
                    ]
                ]
            },
            "properties": {
                "osid": "00000016-e0a2-45ca-855a-645753d72716",
                "toid": "osgb5000005271044468",
                "theme": "Buildings",
                "changetype": "New",
                "isobscured": "false",
                "description": "Building Internal Division",
                "versiondate": "2022-08-26",
                "physicallevel": "Surface Level",
                "geometry_length": 3.734355,
                "geometry_source": "Ordnance Survey",
                "description_source": "Ordnance Survey",
                "geometry_updatedate": "2020-09-23",
                "capturespecification": "Rural",
                "geometry_evidencedate": "2019-07-05",
                "description_updatedate": "2020-09-23",
                "versionavailabletodate": "null",
                "firstdigitalcapturedate": "2020-09-23",
                "description_evidencedate": "2019-07-05",
                "versionavailablefromdate": "2022-08-27T00:00:00Z"
            }
        },
        {
            "id": "00000052-e07e-4420-b2d5-b6ff90afcac2",
            "type": "Feature",
            "geometry": {
                "type": "LineString",
                "coordinates": [
                    [
                        -3.0641844,
                        50.9584926
                    ],
                    [
                        -3.0643669,
                        50.9585053
                    ]
                ]
            },
            "properties": {
                "osid": "00000052-e07e-4420-b2d5-b6ff90afcac2",
                "toid": "osgb1000000283880279",
                "theme": "Buildings",
                "changetype": "New",
                "isobscured": "false",
                "description": "Overhanging Building Edge",
                "versiondate": "2022-08-26",
                "physicallevel": "Level 1",
                "geometry_length": 12.899612,
                "geometry_source": "Ordnance Survey",
                "description_source": "Ordnance Survey",
                "geometry_updatedate": "2004-01-28",
                "capturespecification": "Rural",
                "geometry_evidencedate": "2004-01-28",
                "description_updatedate": "2004-01-28",
                "versionavailabletodate": "null",
                "firstdigitalcapturedate": "1992-02-01",
                "description_evidencedate": "2004-01-28",
                "versionavailablefromdate": "2022-08-27T00:00:00Z"
            }
        },
        {
            "id": "000000b7-a66f-42b5-870b-3ea9729b73e3",
            "type": "Feature",
            "geometry": {
                "type": "LineString",
                "coordinates": [
                    [
                        -1.2987328,
                        54.0082184
                    ],
                    [
                        -1.2987991,
                        54.0081463
                    ]
                ]
            },
            "properties": {
                "osid": "000000b7-a66f-42b5-870b-3ea9729b73e3",
                "toid": "osgb1000000080069326",
                "theme": "Buildings",
                "changetype": "New",
                "isobscured": "false",
                "description": "Building Internal Division",
                "versiondate": "2022-08-26",
                "physicallevel": "Surface Level",
                "geometry_length": 9.130049,
                "geometry_source": "Ordnance Survey",
                "description_source": "Ordnance Survey",
                "geometry_updatedate": "2007-02-20",
                "capturespecification": "Rural",
                "geometry_evidencedate": "2007-02-20",
                "description_updatedate": "2007-02-20",
                "versionavailabletodate": "null",
                "firstdigitalcapturedate": "1992-09-24",
                "description_evidencedate": "2007-02-20",
                "versionavailablefromdate": "2022-08-27T00:00:00Z"
            }
        }
    ]
}


def test_merge_geojsons_pass():
    test_variables = "gj1, gj2, expected_result"
    test_data = [
        param(geojson_1_feature, geojson_3_features, 4),
        param(geojson_1_feature, geojson_10_features, 11),
        param(geojson_3_features, geojson_10_features, 13),
        param(geojson_3_features, {}, 3),
        param({}, geojson_10_features, 10)
    ]
    return test_variables, test_data


def test_merge_geojsons_fail():
    test_variables = "gj1, gj2, expected_result"
    test_data = [
        param({}, {}, ValueError),
        param({"hello": "world"}, {"hello": "world"}, ValueError),
        param({"hello": "world"}, geojson_10_features, ValueError)
    ]
    return test_variables, test_data



import pytest
from osdatahub.utils import clean_polygon

from tests.data import clean_polygon_data as data


class TestCleanPolygon:
    @pytest.mark.parametrize(*data.test_clean_polygon())
    def test_clean_polygon(self, feature, expected_result):
        # Act
        geojson = clean_polygon(feature)

        # Assert
        assert geojson == expected_result
